import asyncio
import os
import sys
from copy import deepcopy
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Callable, Sequence, cast

from pydantic import BaseModel
from typing_extensions import Unpack

from inspect_ai._display import display
from inspect_ai._display._display import TaskProfile
from inspect_ai._util.constants import DEFAULT_EPOCHS
from inspect_ai._util.datetime import iso_now
from inspect_ai._util.dotenv import dotenv_environ
from inspect_ai._util.error import exception_message
from inspect_ai._util.path import chdir_python, cwd_relative_path
from inspect_ai._util.registry import (
    is_registry_object,
    registry_info,
    registry_log_name,
    registry_params,
)
from inspect_ai.dataset import Dataset, MemoryDataset, Sample
from inspect_ai.log import (
    EvalConfig,
    EvalError,
    EvalLog,
    EvalPlan,
    EvalPlanStep,
    EvalStats,
    LoggingMessage,
)
from inspect_ai.log._log import eval_error
from inspect_ai.model import (
    ChatMessage,
    ChatMessageTool,
    ChatMessageUser,
    GenerateConfig,
    GenerateConfigArgs,
    Model,
    ModelName,
    ToolCall,
    ToolFunction,
    ToolInfo,
)
from inspect_ai.model._model import collect_model_usage
from inspect_ai.scorer import Metric, Score, Scorer, Target
from inspect_ai.solver import Generate, Plan, Solver, TaskState, Tool, generate
from inspect_ai.solver._tool.tool import TOOL_PARAMS
from inspect_ai.solver._tool.tool_def import ToolDef, tool_defs
from inspect_ai.util._context.logger import collect_logger_records

from .images import (
    messages_with_base64_images,
    samples_with_base64_images,
)
from .log import EvalLogger
from .score import eval_results, score_async

logger = getLogger(__name__)

TASK_FILE_ATTR = "__task_file__"
TASK_RUN_DIR_ATTR = "__task_run_dir__"


class Task:
    r"""Evaluation task.

    Tasks are the basis for defining and running evaluations. Tasks
    are parameterized with a dataset, a scorer, and metrics. Tasks
    also may optionally provide a default plan for execution.

    Args:
        dataset (Dataset | Sequence[Sample]): Dataset to evaluate
        plan: (Plan | Solver | list[Solver]): Default plan. If not specified
          defaults to generate(), a normal call to the model.
        scorer: (Scorer | None): Scorer used to evaluate model output.
        metrics (list[Metric]): Additional metrics to compute beyond
          the base metrics provided by the scorer.
        config (GenerateConfig): Model generation config.
        epochs (int): Default number of epochs to run for.
        max_messages (int | None): Limit on total messages in the conversation.
        name: (str | None): Task name. If not specified is automatically
          determined based on the name of the task directory (or "task")
          if its anonymous task (e.g. created in a notebook and passed to
          eval() directly)
        version: (int): Version of task (to distinguish evolutions
          of the task spec or breaking changes to it)
    """

    def __init__(
        self,
        dataset: Dataset | Sequence[Sample],
        plan: Plan | Solver | list[Solver] = generate(),
        scorer: Scorer | None = None,
        metrics: list[Metric] = [],
        config: GenerateConfig = GenerateConfig(),
        epochs: int | None = None,
        max_messages: int | None = None,
        name: str | None = None,
        version: int = 0,
    ) -> None:
        self.dataset = (
            dataset if isinstance(dataset, Dataset) else MemoryDataset(list(dataset))
        )
        self.plan = plan if isinstance(plan, Plan) else Plan(plan)
        self.scorer = scorer
        self.metrics = metrics
        self.config = config
        self.epochs = epochs
        self.max_messages = max_messages
        self.version = version
        self._name = name

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name
        elif is_registry_object(self):
            return registry_info(self).name
        else:
            return "task"

    @property
    def attribs(self) -> dict[str, Any]:
        if is_registry_object(self):
            return cast(dict[str, Any], registry_info(self).metadata.get("attribs", {}))
        else:
            return dict()

    async def run(
        self,
        sequence: tuple[int, int],
        model: Model,
        logger: EvalLogger,
        config: EvalConfig = EvalConfig(),
        plan: Plan | Solver | list[Solver] | None = None,
        score: bool = True,
        **kwargs: Unpack[GenerateConfigArgs],
    ) -> EvalLog:
        r"""Run the task.

        Run the task with the passed model and configuration, using the
        samples, scorer, metrics and solver(s) specified for the task.

        Args:
            sequence (int): Sequence of the run within a larger set of runs
            model (Model): Model used to generate output
            logger (EvalLogger): Logger for recording results.
            config (EvalConfig): Config (sample range/epochs, logging options)
            plan:(Plan | Solver | list[Solver] | None): Override of
               task default plan.
            score (bool | None): Score model output. If not specified
              is determined automatically based on whether the task
              has a solver and metrics defined.
            **kwargs (GenerateConfigArgs): Generation config options

        Returns:
          EvalLog for executed task.

        """
        with chdir_python(task_run_dir(self)), dotenv_environ():
            # track stats and error
            stats = EvalStats(started_at=iso_now())
            error: EvalError | None = None

            # see if we are scoring
            score = score and self.scorer is not None

            # evaluate the task (accumulate scores for metrics)
            model_name = ModelName(model)

            # apply limit to dataset
            dataset_limit = (
                slice(0, len(self.dataset))
                if config.limit is None
                else (
                    slice(*config.limit)
                    if isinstance(config.limit, tuple)
                    else slice(0, config.limit)
                )
            )
            dataset = self.dataset[dataset_limit] if dataset_limit else self.dataset

            # add sample ids to dataset if they aren't there (start at 1 not 0)
            for id, sample in zip(
                range(dataset_limit.start, dataset_limit.stop), dataset
            ):
                if sample.id is None:
                    sample.id = id + 1

            # resolve the plan and scorer
            plan = (
                plan
                if isinstance(plan, Plan)
                else Plan(plan)
                if plan is not None
                else self.plan
            )
            scorer: Scorer | None = self.scorer if (score and self.scorer) else None

            # compute the generate() config. we start with the base task config,
            # then merge any deltas provided by the **kwargs for this call to run()
            generate_config = self.config.merge(GenerateConfigArgs(**kwargs))

            # log the plan
            self._log_plan(logger, plan, generate_config)

            # provide solvers a function that they can use to generate output
            async def generate(
                state: TaskState, **kwargs: Unpack[GenerateConfigArgs]
            ) -> TaskState:
                return await self._generate(
                    model=model,
                    state=state,
                    config=generate_config.merge(kwargs),
                    max_messages=config.max_messages,
                )

            # apply epochs (deepcopy the samples so they remain independent)
            epochs = config.epochs if config.epochs else DEFAULT_EPOCHS
            samples: list[Sample] = []
            for _ in range(0, epochs):
                samples.extend([deepcopy(sample) for sample in dataset])

            # if we are logging images then resolve sample images here
            log_images = config.log_images is not False
            if log_images:
                samples = await samples_with_base64_images(samples)

            # prime the eval tasks (deep copy so they share no state w/ sample)
            sample_epochs: list[int] = []
            for e in range(0, epochs):
                sample_epochs.extend([e + 1] * len(dataset))
            states = [
                deepcopy(
                    TaskState(
                        sample_id=sample.id or 0,
                        epoch=epoch,
                        model=model_name,
                        input=sample.input,
                        choices=sample.choices,
                        messages=sample_messages(sample),
                        completed=False,
                        metadata=sample.metadata if sample.metadata else {},
                    )
                )
                for epoch, sample in zip(sample_epochs, samples)
            ]

            # create task profile for display
            profile = TaskProfile(
                name=self.name,
                sequence=sequence,
                model=model_name,
                dataset=self.dataset.name or "(samples)",
                scorer=(
                    registry_log_name(self.scorer)
                    if is_registry_object(self.scorer)
                    else "(none)"
                ),
                samples=len(samples),
                eval_config=config,
                task_args=logger.eval.task_args,
                generate_config=generate_config,
                log_location=logger.location,
            )

            with display().task(profile) as td:
                try:
                    # run w/ progress (steps = samples * steps in plan + 1 for scorer)
                    total_steps = len(samples) * (
                        len(plan.steps) + (1 if plan.finish else 0) + (1)  # scorer
                    )
                    with td.progress(total=total_steps) as p:

                        def progress() -> None:
                            p.update(1)

                        tasks = [
                            self.run_eval_task(
                                sample=sample,
                                state=state,
                                plan=plan,
                                max_messages=config.max_messages,
                                scorer=scorer,
                                generate=generate,
                                progress=progress,
                            )
                            for (sample, state) in zip(samples, states)
                        ]

                        # run them in parallel
                        scores = await asyncio.gather(*tasks)

                    # log output by epoch
                    if config.log_samples is not False:
                        # if we are logging images then be sure to base64 images injected by solvers
                        if log_images:
                            states = await states_with_base64_images(states)

                        for e in range(0, epochs):
                            sl = slice(e * len(dataset), (e + 1) * (len(dataset)))
                            self._log_output(
                                logger, e + 1, samples[sl], states[sl], scores[sl]
                            )

                    # compute and record metrics if we have scores (don't compute metrics on errors)
                    completed_scores = [
                        score for score in scores if isinstance(score, Score)
                    ]
                    if len(completed_scores) > 0:
                        results = eval_results(
                            completed_scores,
                            self.scorer,
                            self.metrics,
                        )
                        logger.log_results(results)

                    # collect eval data
                    collect_eval_data(stats, logger)

                    # display task summary
                    td.summary(results, stats)

                except asyncio.CancelledError as ex:
                    raise ex

                except BaseException as ex:
                    # mark completed
                    stats.completed_at = iso_now()

                    # get exception info
                    type, value, traceback = sys.exc_info()
                    type = type if type else BaseException
                    value = value if value else ex

                    # build eval error
                    error = eval_error(ex, type, value, traceback)

                    # collect eval data
                    collect_eval_data(stats, logger)

                    # display it
                    td.error(error, type, value, traceback)

        # log as appropriate
        if error:
            return logger.log_failure(stats, error)
        else:
            return logger.log_success(stats)

    async def score(self, log: EvalLog) -> EvalLog:
        with chdir_python(task_run_dir(self)), dotenv_environ():
            # confirm we have a scorer
            if self.scorer is None:
                raise ValueError("You must specify a scorer for evals to be scored.")

            # confirm we have samples
            if log.samples is None or len(log.samples) == 0:
                raise ValueError("There are no samples to score in the log.")

            task_name = self.name
            display().print(f"Scoring {len(log.samples)} samples for task: {task_name}")

            # perform scoring
            log = await score_async(log, self.scorer)

        # compute and log metrics
        display().print(f"Aggregating scores for task: {task_name}")
        if self.scorer and log.samples:
            log.results = eval_results(
                [
                    sample.score
                    for sample in log.samples
                    if isinstance(sample.score, Score)
                ],
                self.scorer,
                self.metrics,
            )
        return log

    async def run_eval_task(
        self,
        sample: Sample,
        state: TaskState,
        plan: Plan,
        max_messages: int | None,
        scorer: Scorer | None,
        generate: Generate,
        progress: Callable[..., None],
    ) -> Score | None:
        # solver loop
        try:
            # run plan steps (checking for early termination)
            for index, solver in enumerate(plan.steps):
                # run the solver
                state = await solver(state, generate)
                progress()

                # check for early termination (tick remaining progress)
                if state.completed or has_max_messages(state, max_messages):
                    for _ in range(index + 1, len(plan.steps)):
                        progress()
                    break

            # run finishing step them mark completed
            if plan.finish:
                state = await plan.finish(state, generate)
                progress()
            state.completed = True

        finally:
            # safely run cleanup function if there is one
            if plan.cleanup:
                try:
                    await plan.cleanup(state)
                except Exception as ex:
                    logger.warning(
                        f"Exception occurred during plan cleanup for task {self.name}: "
                        + f"{exception_message(ex)}"
                    )
                    pass

        # score it
        result = await scorer(state, Target(sample.target)) if scorer else None
        progress()

        # return
        return result

    async def _generate(
        self,
        model: Model,
        state: TaskState,
        config: GenerateConfig,
        max_messages: int | None,
    ) -> TaskState:
        # track tool_choice (revert to "none" after first forced call of a tool)
        tool_choice = state.tool_choice

        while True:
            # call the model
            output = await model.generate(
                state.messages,
                tools_info(state.tools),
                tool_choice,
                config,
            )

            # append the assistant message
            message = output.choices[0].message
            state.messages.append(message)

            # check for max messages
            if has_max_messages(state, max_messages):
                state.output = output
                return state

            # resolve tool calls if necessary
            tdefs = tool_defs(state.tools)
            if message.tool_calls and len(message.tool_calls) > 0:
                for tool_call in message.tool_calls:
                    tool_error: str | None = None
                    try:
                        result = await call_tool(tdefs, tool_call, state.metadata)
                    except Exception as ex:
                        result = ""
                        tool_error = exception_message(ex)

                    if isinstance(result, tuple):
                        result, metadata = result
                        state.metadata.update(metadata)

                    state.messages.append(
                        ChatMessageTool(
                            content=str(result),
                            tool_error=tool_error,
                            tool_call_id=tool_call.id,
                        )
                    )

                    # check for max messages
                    if has_max_messages(state, max_messages):
                        state.output = output
                        return state

                    # if a tool_call was forced set tool_choice to 'none'
                    # (otherwise it will get forced over and over again)
                    if isinstance(tool_choice, ToolFunction):
                        tool_choice = "none"

            # no tool calls, we are done!
            else:
                state.output = output
                return state

    def _log_output(
        self,
        logger: EvalLogger,
        epoch: int,
        samples: list[Sample],
        states: list[TaskState],
        scores: list[Score | None],
    ) -> None:
        for i in range(len(samples)):
            logger.log_sample(epoch, samples[i], states[i], scores[i])

    def _log_plan(
        self,
        logger: EvalLogger,
        plan: Plan,
        config: GenerateConfig,
    ) -> None:
        def eval_plan_step(solver: Solver) -> EvalPlanStep:
            return EvalPlanStep(
                solver=registry_log_name(solver), params=registry_params(solver)
            )

        eval_plan = EvalPlan(
            name=plan.name,
            steps=[eval_plan_step(solver) for solver in plan.steps],
            finish=eval_plan_step(plan.finish) if plan.finish else None,
            config=config,
        )
        if plan.finish:
            eval_plan.steps.append(eval_plan_step(plan.finish))

        logger.log_event("plan", eval_plan)


class TaskInfo(BaseModel):
    """Task information (file, name, and attributes)."""

    file: str
    """File path where task was loaded from."""

    name: str
    """Task name (defaults to function name)"""

    attribs: dict[str, Any]
    """Task attributes (arguments passed to `@task`)"""

    def __str__(self) -> str:
        return f"{self.file}@{self.name}"

    def __hash__(self) -> int:
        return hash(
            (self.file, self.name)
            + tuple(self.attribs.keys())
            + tuple(self.attribs.values())
        )


@dataclass
class TaskSpec:
    id: str
    task: str


Tasks = (
    str
    | TaskSpec
    | TaskInfo
    | Task
    | Callable[..., Task]
    | type[Task]
    | list[str]
    | list[TaskInfo]
    | list[Task]
    | list[Callable[..., Task]]
    | list[type[Task]]
    | None
)
r"""One or more tasks.

Tasks to be evaluated. Many forms of task specification are
supported including directory names, task functions, task
classes, and task instances (a single task or list of tasks
can be specified). None is a request to read a task out
of the current working directory.
"""


def task_file(task: Task, relative: bool = False) -> str | None:
    file = cast(str | None, getattr(task, TASK_FILE_ATTR, None))
    if file:
        if relative:
            return cwd_relative_path(file)
        else:
            return file
    else:
        return None


def task_run_dir(task: Task) -> str:
    return getattr(task, TASK_RUN_DIR_ATTR, os.getcwd())


def sample_messages(sample: Sample) -> list[ChatMessage]:
    if isinstance(sample.input, str):
        return [ChatMessageUser(content=sample.input, source="input")]
    else:
        messages = deepcopy(sample.input)
        for message in messages:
            message.source = "input"
        return messages


def has_max_messages(state: TaskState, max_messages: int | None) -> bool:
    return max_messages is not None and (len(state.messages) >= max_messages)


async def states_with_base64_images(states: list[TaskState]) -> list[TaskState]:
    return await asyncio.gather(*[state_with_base64_images(state) for state in states])


async def state_with_base64_images(state: TaskState) -> TaskState:
    state.messages = await messages_with_base64_images(state.messages)
    return state


def collect_eval_data(stats: EvalStats, logger: EvalLogger) -> None:
    # collect stats
    stats.completed_at = iso_now()
    stats.model_usage = collect_model_usage()

    # collect log output
    for record in collect_logger_records():
        logger.log_event("logging", LoggingMessage.from_log_record(record))


def tools_info(tools: list[Tool]) -> list[ToolInfo]:
    tdefs = tool_defs(tools)
    return [
        ToolInfo(name=tool.name, description=tool.description, params=tool.params)
        for tool in tdefs
    ]


async def call_tool(
    tools: list[ToolDef], call: ToolCall, metadata: dict[str, Any]
) -> Any:
    # find the tool
    tool_def = next((tool for tool in tools if tool.name == call.function), None)
    if tool_def is None:
        return f"Tool {call.function} not found"

    # resolve metadata params and prepend to arguments
    tool_params: dict[str, str] = registry_info(tool_def.tool).metadata.get(
        TOOL_PARAMS, {}
    )
    resolved_params: dict[str, Any] = {}
    for name, value in tool_params.items():
        key = value.removeprefix("metadata.")
        resolved = metadata.get(key, None)
        if resolved is None:
            raise ValueError(f"Metadata value '{key}' not found for tool parameter")
        resolved_params[name] = resolved
    arguments = resolved_params | call.arguments

    # call the tool
    try:
        return await tool_def.tool(**arguments)
    except Exception as e:
        return f"Error: {exception_message(e)}"



import json
from io import TextIOWrapper
from pathlib import Path
from typing import Any, cast

import jsonlines

from inspect_ai._util.file import file

from .._dataset import (
    Dataset,
    DatasetReader,
    FieldSpec,
    MemoryDataset,
    RecordToSample,
)
from .._util import record_to_sample_fn


def json_dataset(
    json_file: str,
    sample_fields: FieldSpec | RecordToSample | None = None,
    shuffle: bool = False,
    seed: int | None = None,
    limit: int | None = None,
    encoding: str = "utf-8",
    name: str | None = None,
    fs_options: dict[str, Any] = {},
) -> Dataset:
    r"""Read dataset from a JSON file.

    Read a dataset from a JSON file containing an array of objects, or
    from a JSON Lines file containing one object per line. These objects may
    already be formatted as `Sample` instances, or may require some mapping using
    the `sample_fields` argument.

    Args:
      json_file (str): Path to JSON file. Can be a local filesystem path or
        a path to an S3 bucket (e.g. "s3://my-bucket"). Use `fs_options`
        to pass arguments through to the `S3FileSystem` constructor.
      sample_fields (SampleFieldSpec | RecordToSample): Method of mapping underlying
        fields in the data source to `Sample` objects. Pass `None` if the data is already
        stored in `Sample` form (i.e. object with "input" and "target" fields); Pass a
        `SampleFieldSpec` to specify mapping fields by name; Pass a `RecordToSample` to
        handle mapping with a custom function.
      shuffle (bool): Randomly shuffle the dataset order.
      seed: (int | None): Seed used for random shuffle.
      limit (int | None): Limit the number of records to read.
      encoding (str): Text encoding for file (defaults to "utf-8").
      name (str): Optional name for dataset (for logging). If not specified,
        defaults to the stem of the filename.
      fs_options (dict[str, Any]): Optional. Addional arguments to pass through
        to the filesystem provider (e.g. `S3FileSystem`). Use `{"anon": True }`
        if you are accessing a public S3 bucket with no credentials.

    Returns:
        Dataset read from JSON file.
    """
    # resolve data_to_sample function
    data_to_sample = record_to_sample_fn(sample_fields)

    # pick the right reader for the file extension
    dataset_reader = (
        jsonlines_dataset_reader
        if json_file.lower().endswith(".jsonl")
        else json_dataset_reader
    )

    # read and convert samples
    with file(json_file, "r", encoding=encoding, fs_options=fs_options) as f:
        name = name if name else Path(json_file).stem
        dataset = MemoryDataset(
            samples=[data_to_sample(data) for data in dataset_reader(f)],
            name=name,
            location=json_file,
        )

        # shuffle if requested
        if shuffle:
            dataset.shuffle(seed=seed)

        # limit if requested
        if limit:
            dataset = MemoryDataset(list(dataset[0:limit]))

    return dataset


def jsonlines_dataset_reader(file: TextIOWrapper) -> DatasetReader:
    jsonlines_reader = jsonlines.Reader(file)
    return jsonlines_reader.iter(type=dict)


def json_dataset_reader(file: TextIOWrapper) -> DatasetReader:
    data = cast(list[dict[str, Any]], json.load(file))
    return iter(data)



import re


def is_http_url(url: str) -> bool:
    return url.startswith("http://") or url.startswith("https://")


def is_data_uri(url: str) -> bool:
    return url.startswith("data:")


def data_uri_mime_type(data_url: str) -> str | None:
    pattern = r"^data:([^;]+);.*"
    match = re.match(pattern, data_url)
    if match:
        mime_type = match.group(1)
        return mime_type
    else:
        return None


def data_uri_to_base64(data_uri: str) -> str:
    pattern = r"^data:[^,]+,"
    stripped_uri = re.sub(pattern, "", data_uri)
    return stripped_uri



import ast
import inspect
import os
import re
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader
from logging import getLogger
from pathlib import Path
from types import ModuleType
from typing import Any, Callable

from inspect_ai._util.dotenv import dotenv_environ
from inspect_ai._util.error import exception_message
from inspect_ai._util.file import file
from inspect_ai._util.path import chdir_python
from inspect_ai._util.registry import RegistryInfo, is_registry_object, registry_info
from inspect_ai.model import ModelName

from .registry import task_create
from .task import TASK_FILE_ATTR, TASK_RUN_DIR_ATTR, Task, TaskInfo

logger = getLogger(__name__)


def list_tasks(
    globs: str | list[str] = [],
    absolute: bool = False,
    root_dir: Path = Path.cwd(),
    filter: Callable[[TaskInfo], bool] | None = None,
) -> list[TaskInfo]:
    """List the tasks located at the specified locations.

    Args:
        globs (str | list[str]): File location(s). Can be
           globs (e.g. have bash-style wildcards).
        absolute (bool): Return absolute paths (defaults
           to False)
        root_dir (Path): Base directory to scan from
           (defaults to current working directory)
        filter (Callable[[TaskInfo], bool] | None):
           Filtering function.

    Returns:
        List of TaskInfo
    """
    # resovle globs
    globs = globs if isinstance(globs, list) else [globs]

    # build list of tasks to return
    tasks: list[TaskInfo] = []
    files = task_files(globs, root_dir)
    for task_file in files:
        tasks.extend(parse_tasks(task_file, root_dir, absolute))

    # filter if necessary
    tasks = [task for task in tasks if filter is None or filter(task)]

    # return sorted
    return sorted(tasks, key=lambda t: f"{t.file}@{t.name}")


def create_tasks(
    globs: list[str],
    model: ModelName,
    task_args: dict[str, Any] = {},
    root_dir: Path | None = None,
) -> list[Task]:
    tasks: list[Task] = []

    root_dir = root_dir if root_dir is not None else Path.cwd()

    for glob in globs:
        # sometimes globs are direct references to files
        # that inclue an @ index. for this case directly
        # create the task (we also need to load the file
        # so the task is registered before we create it)
        spec_split = split_task_spec(glob)
        if len(spec_split[1]) > 0:
            task_path = Path(spec_split[0])
            load_file_tasks(task_path.absolute())
            tasks.extend(
                create_file_tasks(task_path, model, [spec_split[1]], task_args)
            )
        else:
            # if the glob is the root dir then set it to empty (will result in
            # enumeration of the root dir)
            target = [] if Path(glob).resolve() == root_dir.resolve() else [glob]
            files = task_files(target, root_dir)
            files = sorted(files, key=lambda f: f.as_posix())
            for file in files:
                tasks.extend(create_file_tasks(file, model, None, task_args))
    return tasks


def task_files(globs: list[str] = [], root_dir: Path | None = None) -> list[Path]:
    # root dir
    root_dir = root_dir if root_dir else Path.cwd()

    # no globs is cwds
    if len(globs) == 0:
        return tasks_in_dir(root_dir)

    # resolve the first level of globs
    paths: list[Path] = []
    for glob in globs:
        # we will have matched a set of directories and files
        # (depending on how the user wrote the globs). for
        # each file, add it to to our list if its a task file;
        # for each dir, recursively search it for task files
        expanded = list(root_dir.glob(glob))
        for path in expanded:
            if path.is_dir():
                paths.extend(tasks_in_dir(path))
            elif is_task_path(path):
                paths.append(path)

    return [path.absolute() for path in paths]


def tasks_in_dir(path: Path) -> list[Path]:
    paths: list[Path] = []
    for dir, dirnames, filenames in os.walk(path):
        # compute dir_path
        dir_path = Path(dir)

        # remove dirs that start with . or _
        dirnames[:] = [
            dirname for dirname in dirnames if not is_task_path_excluded(dirname)
        ]

        # select files w/ the right extension
        for filename in filenames:
            file_path = dir_path / filename
            if is_task_path(file_path):
                paths.append(file_path)

    return paths


def load_file_tasks(file: Path) -> list[RegistryInfo]:
    with chdir_python(file.parent.as_posix()), dotenv_environ():
        return _load_task_specs(file)


def create_file_tasks(
    file: Path,
    model: ModelName,
    task_specs: list[str] | list[RegistryInfo] | None = None,
    task_args: dict[str, Any] = {},
) -> list[Task]:
    with chdir_python(file.parent.as_posix()), dotenv_environ():
        # if we don't have task specs then go get them (also,
        # turn them into plain names)
        if task_specs is None:
            task_specs = _load_task_specs(file)
        # convert to plain names
        task_specs = [
            spec if isinstance(spec, str) else spec.name for spec in task_specs
        ]

        tasks: list[Task] = []
        for task_spec in task_specs:
            # create the task from the loaded source file and
            # note that it was loaded from this directory
            # (will be used later to ensure it runs in the directory)
            task = task_create(task_spec, model, **task_args)
            setattr(task, TASK_FILE_ATTR, file.as_posix())
            setattr(task, TASK_RUN_DIR_ATTR, file.parent.as_posix())
            tasks.append(task)
        return tasks


# don't call this function directly, rather, call one of the
# higher level loading functions above (those functions
# change the working directory, this one does not b/c it is
# intended as a helper funciton)
def _load_task_specs(task_path: Path) -> list[RegistryInfo]:
    # load the module
    module = load_task_module(task_path)
    if module:
        # find the tasks in the module
        tasks = inspect.getmembers(module, lambda m: is_registry_object(m, "task"))
        return [registry_info(task[1]) for task in tasks]
    else:
        return []


excluded_pattern = re.compile("^[_\\.].*$")


def is_task_path_excluded(path: str) -> bool:
    return (
        re.match(excluded_pattern, path) is not None
        or path == "env"
        or path == "venv"
        or path == "tests"
    )


def is_task_path(path: Path) -> bool:
    return (
        path.suffix == ".py" or path.suffix == ".ipynb"
    ) and not is_task_path_excluded(path.name)


def split_task_spec(task_spec: str) -> tuple[str, str]:
    parts = task_spec.rsplit("@", 1)
    if len(parts) == 2:
        return parts[0], parts[1]
    else:
        return task_spec, ""


def load_task_module(task_path: Path) -> ModuleType | None:
    if task_path.suffix == ".py":
        # bail if the code doesn't have a task
        with open(task_path, "r", encoding="utf-8") as file:
            if not code_has_task(file.read()):
                return None

        module_name = task_path.as_posix()
        loader = SourceFileLoader(module_name, task_path.absolute().as_posix())
        spec = spec_from_loader(loader.name, loader)
        if not spec:
            raise ModuleNotFoundError(f"Module {module_name} not found")
        module = module_from_spec(spec)
        loader.exec_module(module)
        return module

    elif task_path.suffix == ".ipynb":
        try:
            from inspect_ai._util.notebook import NotebookLoader
        except ImportError:
            return None

        # bail if the code doesn't have a task
        def exec_filter(cells: list[str]) -> bool:
            code = "\n\n".join(cells)
            return code_has_task(code)

        notebook_loader = NotebookLoader(exec_filter)
        return notebook_loader.load_module(task_path.as_posix())

    else:
        raise ModuleNotFoundError(
            f"Invalid extension for task file: {task_path.suffix}"
        )


def code_has_task(code: str) -> bool:
    try:
        tree = ast.parse(code)
        for node in ast.iter_child_nodes(tree):
            if isinstance(node, ast.FunctionDef):
                for decorator in node.decorator_list:
                    if isinstance(decorator, ast.Name):
                        if str(decorator.id) == "task":
                            return True
                    elif isinstance(decorator, ast.Call):
                        if isinstance(decorator.func, ast.Name):
                            if str(decorator.func.id) == "task":
                                return True
    except SyntaxError:
        pass

    return False


def parse_tasks(path: Path, root_dir: Path, absolute: bool) -> list[TaskInfo]:
    # read code from python source file
    if path.suffix.lower() == ".py":
        with file(path.as_posix(), "r", encoding="utf-8") as f:
            code = f.read()

    # read code from notebook
    elif path.suffix.lower() == ".ipynb":
        try:
            from inspect_ai._util.notebook import read_notebook_code
        except ImportError:
            return []

        code = read_notebook_code(path)

    # unsupported file type
    else:
        raise ModuleNotFoundError(f"Invalid extension for task file: {path.suffix}")

    # parse the top level tasks out of the code
    tasks: list[TaskInfo] = []
    try:
        tree = ast.parse(code)
        for node in ast.iter_child_nodes(tree):
            if isinstance(node, ast.FunctionDef):
                for decorator in node.decorator_list:
                    result = parse_decorator(node, decorator)
                    if result:
                        name, attribs = result
                        tasks.append(
                            TaskInfo(
                                file=task_path(path, root_dir, absolute),
                                name=name,
                                attribs=attribs,
                            )
                        )
    except SyntaxError:
        pass

    return tasks


def parse_decorator(
    node: ast.FunctionDef, decorator: ast.expr
) -> tuple[str, dict[str, Any]] | None:
    if isinstance(decorator, ast.Name):
        if str(decorator.id) == "task":
            return node.name, {}
    elif isinstance(decorator, ast.Call):
        if isinstance(decorator.func, ast.Name):
            if str(decorator.func.id) == "task":
                return parse_task_decorator(node, decorator)
    return None


def parse_task_decorator(
    node: ast.FunctionDef, decorator: ast.Call
) -> tuple[str, dict[str, Any]]:
    name = node.name
    attribs: dict[str, Any] = {}
    for arg in decorator.keywords:
        if arg.arg is not None:
            try:
                value = ast.literal_eval(arg.value)
                if arg.arg == "name":
                    name = value
                else:
                    attribs[arg.arg] = value
            except ValueError as ex:
                # when parsing tasks, we can't provide the values of expressions that execute code
                logger.debug(
                    f"Error parsing attribute {arg.arg} of task {node.name}: {exception_message(ex)}"
                )
                pass
    return name, attribs


# manage relative vs. absolute paths
def task_path(path: Path, root_dir: Path, absolute: bool) -> str:
    if absolute:
        return path.resolve().as_posix()
    else:
        return path.relative_to(root_dir.resolve()).as_posix()



import re

from inspect_ai.solver import TaskState

from ._metric import CORRECT, INCORRECT, Score
from ._metrics import accuracy, bootstrap_std
from ._scorer import Scorer, Target, scorer


@scorer(metrics=[accuracy(), bootstrap_std()])
def pattern(pattern: str, ignore_case: bool = True) -> Scorer:
    """Scorer which extracts the model answer using a regex.

    The regex can have a single capture group or multiple
    groups. In the case of multiple groups, the first
    group is bypassed (as the prefix of the answer) and
    the second group is used for the answer.

    Args:
       pattern (str): Regular expression for extracting the
          answer from model output.
       ignore_case (bool): Ignore case when comparing
          the extract answer to the targets.
    """

    async def score(state: TaskState, target: Target) -> Score:
        # extract the answer
        match = re.search(
            pattern, state.output.completion, re.IGNORECASE if ignore_case else 0
        )

        # got a match
        if match:
            # handle case insentitive
            answer = match.group(1) if len(match.groups()) == 1 else match.group(2)
            input = answer
            if ignore_case:
                input = input.lower()
                target = Target([t.lower() for t in target])

            # return score
            return Score(
                value=CORRECT if input in target else INCORRECT,
                answer=answer,
                explanation=state.output.completion,
            )
        # didn't find the scoring pattern
        else:
            return Score(
                value=INCORRECT,
                explanation="Scoring pattern not matched in output: "
                + f"{state.output.completion}",
            )

    return score



import pytest
from utils import skip_if_no_cloudflare

from inspect_ai.model import get_model


@pytest.mark.asyncio
@skip_if_no_cloudflare
async def test_cloudflare_api() -> None:
    model = get_model("cf/meta/llama-2-7b-chat-fp16")
    message = "This is a test string. What are you?"
    response = await model.generate(input=message)
    assert len(response.completion) >= 1



"""
HellaSwag: Can a Machine Really Finish Your Sentence?

Rowan Zellers, Ari Holtzman, Yonatan Bisk, Ali Farhadi, Yejin Choi
https://arxiv.org/abs/1905.07830
"""

from inspect_ai import Task, task
from inspect_ai.dataset import Sample, hf_dataset
from inspect_ai.scorer import answer
from inspect_ai.solver import multiple_choice, system_message

SYSTEM_MESSAGE = """
Choose the most plausible continuation for the story.
"""


def record_to_sample(record):
    return Sample(
        input=record["ctx"],
        target=chr(ord("A") + int(record["label"])),
        choices=record["endings"],
        metadata=dict(source_id=record["source_id"]),
    )


@task
def hellaswag():
    # dataset
    dataset = hf_dataset(
        path="hellaswag",
        split="validation",
        sample_fields=record_to_sample,
        trust=True,
        shuffle=True,
    )

    # define task
    return Task(
        dataset=dataset,
        plan=[system_message(SYSTEM_MESSAGE), multiple_choice()],
        scorer=answer("letter"),
    )



from json import dumps

import click

from inspect_ai import __version__
from inspect_ai._util.constants import PKG_PATH
from inspect_ai.log import eval_log_json, read_eval_log


@click.group("info")
def info_command() -> None:
    """Read configuration and log info."""
    return None


@info_command.command("version")
@click.option(
    "--json",
    type=bool,
    is_flag=True,
    default=False,
    help="Output version and path info as JSON",
)
def version(json: bool) -> None:
    if json:
        print(dumps(dict(version=__version__, path=PKG_PATH.as_posix()), indent=2))
    else:
        print(f"version: {__version__}")
        print(f"path: {PKG_PATH.as_posix()}")


@info_command.command("log-file")
@click.argument("path")
@click.option(
    "--header-only",
    type=bool,
    is_flag=True,
    default=False,
    help="Read and print only the header of the log file (i.e. no samples).",
)
def log(path: str, header_only: bool) -> None:
    """Print log file contents."""
    log = read_eval_log(path, header_only=header_only)
    print(eval_log_json(log))


@info_command.command("log-schema")
def log_schema() -> None:
    """Print JSON schema for log files."""
    print(view_resource("log-schema.json"))


@info_command.command("log-types")
def log_types() -> None:
    """Print TS declarations for log files."""
    print(view_resource("log.d.ts"))


def view_resource(file: str) -> str:
    resource = PKG_PATH / "_view" / "www" / file
    with open(resource, "r", encoding="utf-8") as f:
        return f.read()



import pytest
from transformers import PreTrainedModel  # type: ignore
from utils import skip_if_github_action

from inspect_ai.model import (
    ChatMessageUser,
    GenerateConfig,
    get_model,
)


@pytest.fixture
def model() -> PreTrainedModel:
    return get_model(
        "hf/EleutherAI/pythia-70m",
        config=GenerateConfig(
            max_tokens=1,
            seed=42,
            temperature=0.01,
        ),
        # this allows us to run base models with the chat message scaffolding:
        chat_template="{% for message in messages %}{{ message.content }}{% endfor %}",
    )


@pytest.mark.asyncio
@skip_if_github_action
async def test_hf_api(model: PreTrainedModel) -> None:
    message = ChatMessageUser(content="Lorem ipsum dolor")
    response = await model.generate(input=[message])
    assert len(response.completion) >= 1


@pytest.mark.asyncio
@skip_if_github_action
async def test_hf_api_fails(model: PreTrainedModel) -> None:
    temp_before = model.config.temperature
    try:
        model.config.temperature = 0.0

        message = ChatMessageUser(content="Lorem ipsum dolor")
        with pytest.raises(Exception):
            await model.generate(input=[message])
    finally:
        model.config.temperature = temp_before



import os

from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_TOKENS
from inspect_ai.model._providers.util import model_base_url

from .._model import GenerateConfig
from .openai import OpenAIAPI


class TogetherAIAPI(OpenAIAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
    ) -> None:
        api_key = os.environ.get("TOGETHER_API_KEY", None)
        if not api_key:
            raise RuntimeError("TOGETHER_API_KEY environment variable not set")
        base_url = model_base_url(base_url, "TOGETHER_BASE_URL")
        base_url = base_url if base_url else "https://api.together.xyz/v1"
        super().__init__(
            model_name=model_name, base_url=base_url, config=config, api_key=api_key
        )

    # Together uses a default of 512 so we bump it up
    @override
    def max_tokens(self) -> int:
        return DEFAULT_MAX_TOKENS



import numpy as np

from .._metric import Metric, Score, metric


@metric
def mean() -> Metric:
    """Compute mean of all scores.

    Returns:
       mean metric
    """

    def metric(scores: list[Score]) -> float:
        return np.mean([score.as_float() for score in scores]).item()

    return metric


@metric
def var() -> Metric:
    """Compute variance over all scores.

    Returns:
       var metric
    """

    def metric(scores: list[Score]) -> float:
        return np.var([score.as_float() for score in scores]).item()

    return metric



import asyncio
from contextvars import ContextVar
from dataclasses import dataclass


def concurrency(
    name: str,
    concurrency: int,
    key: str | None = None,
) -> asyncio.Semaphore:
    """Obtain a concurrency context.

    A concurrency context can be used to limit the number of coroutines
    executing a block of code (e.g calling an API). For example, here
    we limit concurrent calls to an api ('api-name') to 10:

    ```python
    async with concurrency("api-name", 10):
        # call the api
    ```

    Note that concurrency for model API access is handled internally
    via the `max_connections` generation config option. Concurrency
    for launching subprocesses is handled via the `subprocess` function.

    Args:
      name (str): Name for concurrency context. This serves as the
         display name for the context, and also the unique context
         key (if the `key` parameter is ommitted)
      concurrency (int): Maximum number of couroutines that can
         enter the context.
      key (str | None): Unique context key for this context. Optional.
         Used if the unique key isn't human readable -- e.g. includes
         api tokens or account ids so that the more readable `name`
         can be presented to users e.g in console UI>

    Returns:
       Asyncio Semaphore for concurrency context.
    """
    # sort out key
    key = key if key else name

    # get semaphores dict (only valid when an eval is running)
    concurrency_semaphores = concurrency_semaphores_context_var.get(None)
    if concurrency_semaphores is None:
        raise RuntimeError("Attempted to get eval sempahore when eval not running")

    # do we have an existing semaphore? if not create one and store it
    semaphore = concurrency_semaphores.get(key, None)
    if semaphore is None:
        semaphore = ConcurencySempahore(
            name, concurrency, asyncio.Semaphore(concurrency)
        )
        concurrency_semaphores[key] = semaphore

    # return the semaphore
    return semaphore.semaphore


def init_concurrency() -> None:
    concurrency_semaphores_context_var.set({})


def using_concurrency() -> bool:
    return concurrency_semaphores_context_var.get(None) is not None


def concurrency_status() -> dict[str, tuple[int, int]]:
    if using_concurrency():
        status: dict[str, tuple[int, int]] = {}
        for c in concurrency_semaphores_context_var.get().values():
            status[c.name] = (c.concurrency - c.semaphore._value, c.concurrency)
        return status
    else:
        return {}


@dataclass
class ConcurencySempahore:
    name: str
    concurrency: int
    semaphore: asyncio.Semaphore


concurrency_semaphores_context_var = ContextVar[dict[str, ConcurencySempahore]](
    "concurrency_sempahores"
)



import asyncio
import os
from typing import Any, Literal, Protocol, cast, runtime_checkable

import httpx
from bs4 import BeautifulSoup, NavigableString

from inspect_ai.model import Model, get_model
from inspect_ai.util import concurrency

from .tool import Tool, tool

DEFAULT_RELEVANCE_PROMPT = """I am trying to answer the following question and need to find the most relevant information on the web. Please let me know if the following content is relevant to the question or not. You should just respond with "yes" or "no".

Question: {question}
Page Content: {text}
"""


@tool(
    prompt="""Please use web search to assist in answering the question. If you already know the answer, you do not need to use this tool. If the search results are not helpful, please just take your best guess."""
)
def web_search(
    provider: Literal["google"] = "google",
    num_results: int = 3,
    max_provider_calls: int = 3,
    max_connections: int = 10,
    model: str | Model | None = None,
) -> Tool:
    """Web search tool.

    A tool that can be registered for use by models to search the web. Use
    the `use_tools()` solver to make the tool available (e.g. `use_tools(web_search())`))

    A web search is conducted using the specified provider, the results are parsed for relevance
    using the specified model, and the top 'num_results' relevant pages are returned.

    Args:
      provider (Literal["google"]): Search provider (defaults to "google", currently
        the only provider). Possible future providers include "brave" and "bing".
      num_results (int): Number of web search result pages to return to the model.
      max_provider_calls (int): Maximum number of search calls to make to the search provider.
      max_connections (int): Maximum number of concurrent connections to API
        endpoint of search provider.
      model (str | Model): Model used to parse web pages for relevance.

    Returns:
       A tool that can be registered for use by models to search the web.
    """
    # get search client
    client = httpx.AsyncClient()

    # resolve provider (only google for now)
    if provider == "google":
        search_provider = google_search_provider(client)
    else:
        raise ValueError(f"Unsupported search provider: {provider}")

    # resolve model
    relevance_model = get_model(model)

    async def execute(query: str) -> tuple[str, dict[str, Any]]:
        """
        Tool for searching the web.

        Args:
            query (str): Search query.
        """
        # limit number of concurrent searches
        page_contents: list[str] = []
        urls: list[str] = []
        snippets: list[str] = []
        search_calls = 0

        # Paginate through search results until we have successfully extracted num_results pages or we have reached max_provider_calls
        while len(page_contents) < num_results and search_calls < max_provider_calls:
            async with concurrency(f"{provider}_web_search", max_connections):
                links = await search_provider(query, start_idx=search_calls * 10)

            # Extract and summarize each page individually
            pages = await asyncio.gather(
                *[
                    page_if_relevant(link.url, query, relevance_model, client)
                    for link in links
                ],
                return_exceptions=True,
            )
            for page, link in zip(pages, links):
                if page and not isinstance(page, Exception):
                    page_contents.append(cast(str, page))
                    urls.append(link.url)
                    snippets.append(link.snippet)
            search_calls += 1

        all_page_contents = "\n\n".join(page_contents)
        if all_page_contents == "":
            response = "I'm sorry, I couldn't find any relevant information on the web."
        else:
            response = (
                "Here are your web search results. Please read them carefully as they may be useful later! "
                + all_page_contents
            )

        results = [
            dict(
                url=url,
                snippet=snippet,
            )
            for url, snippet in zip(urls, snippets)
        ]
        return response, {"web_search": {"query": query, "results": results}}

    return execute


async def page_if_relevant(
    link: str, query: str, relevance_model: Model, client: httpx.AsyncClient
) -> str | None:
    """
    Use parser model to determine if a web page contents is relevant to a query.

    Args:
        link (str): Web page link.
        query (str): Search query.
        relevance_model (Model): Model used to parse web pages for relevance.
        client: (httpx.Client): HTTP client to use to fetch the page

    Returns:
        str: Web page contents if relevant, else None.
    """
    # retreive document
    try:
        response = await client.get(link)
        response.raise_for_status()
    except httpx.HTTPError as exc:
        raise Exception(f"HTTP error occurred: {exc}")

    # parse it
    encoding_scheme = response.encoding or "utf-8"
    soup = BeautifulSoup(response.content.decode(encoding_scheme), "html.parser")

    main_content = soup.find("main") or soup.find("body") or soup
    if not isinstance(main_content, NavigableString):
        paragraphs = main_content.find_all("p")
        full_text = ""
        for p in paragraphs:
            full_text += p.get_text(strip=True, separator=" ")
            if len(full_text.split()) > 2000:
                break
    else:
        full_text = " ".join(
            main_content.get_text(strip=True, separator=" ").split()[:2000]
        )

    is_relevant = (
        (
            await relevance_model.generate(
                DEFAULT_RELEVANCE_PROMPT.format(question=query, text=full_text)
            )
        )
        .choices[0]
        .message.text
    )

    if "yes" in is_relevant.lower():
        return full_text
    else:
        return None


class SearchLink:
    def __init__(self, url: str, snippet: str) -> None:
        self.url = url
        self.snippet = snippet


@runtime_checkable
class SearchProvider(Protocol):
    async def __call__(self, query: str, start_idx: int) -> list[SearchLink]: ...


def google_search_provider(client: httpx.AsyncClient) -> SearchProvider:
    google_api_key = os.environ.get("GOOGLE_CSE_API_KEY", None)
    google_cse_id = os.environ.get("GOOGLE_CSE_ID", None)
    if not google_api_key or not google_cse_id:
        raise Exception(
            "GOOGLE_CSE_ID and/or GOOGLE_CSE_API_KEY not set in environment"
        )

    async def search(query: str, start_idx: int) -> list[SearchLink]:
        # List of allowed parameters can be found https://developers.google.com/custom-search/v1/reference/rest/v1/cse/list
        search_params = {
            "q": query,
            "key": google_api_key,
            "cx": google_cse_id,
            "start": start_idx,
        }
        search_url = "https://www.googleapis.com/customsearch/v1?" + "&".join(
            [f"{key}={value}" for key, value in search_params.items()]
        )
        result = await client.get(search_url)
        data = result.json()
        if "items" in data:
            return [SearchLink(item["link"], item["snippet"]) for item in data["items"]]
        else:
            return []

    return search



import os
from json import dumps
from typing import Literal
from urllib.parse import urlparse

import click
from fsspec.core import split_protocol  # type: ignore
from pydantic_core import to_jsonable_python
from typing_extensions import Unpack

from inspect_ai._cli.common import CommonOptions, common_options, resolve_common_options
from inspect_ai._cli.util import parse_cli_args
from inspect_ai._eval.list import list_tasks
from inspect_ai._eval.task import TaskInfo
from inspect_ai.log import list_eval_logs


@click.group("list")
def list_command() -> None:
    """List tasks or eval logs."""
    return None


@list_command.command()
@click.option(
    "-F",
    multiple=True,
    type=str,
    help="One or more boolean task filters (e.g. -F light=true or -F draft~=false)",
)
@click.option(
    "--absolute",
    type=bool,
    is_flag=True,
    default=False,
    help="List absolute paths to task scripts (defaults to relative to the cwd).",
)
@click.option(
    "--json",
    type=bool,
    is_flag=True,
    default=False,
    help="Output listing as JSON",
)
@click.argument("paths", nargs=-1)
@common_options
def tasks(
    paths: tuple[str] | None,
    f: tuple[str] | None,
    absolute: bool,
    json: bool,
    **kwargs: Unpack[CommonOptions],
) -> None:
    """List tasks in given directories."""
    # resolve common options
    resolve_common_options(kwargs)

    # parse filter expressions and build a filter from it
    filters = parse_cli_args(f)

    def task_filter(task: TaskInfo) -> bool:
        for name, value in filters.items():
            if name.endswith("~"):
                name = name[:-1]
                include = task.attribs.get(name, None) != value
            else:
                include = task.attribs.get(name, None) == value
            if not include:
                return False
        return True

    # list tasks
    tasks = list_tasks(
        globs=list(paths) if paths else [], absolute=absolute, filter=task_filter
    )

    # print as JSON or plain text
    if json:
        print(dumps(to_jsonable_python(tasks), indent=2))
    else:
        print("\n".join([f"{task.file}@{task.name}" for task in tasks]))


@list_command.command()
@click.option(
    "--status",
    type=click.Choice(["started", "success", "error"], case_sensitive=False),
    help="List only log files with the indicated status.",
)
@click.option(
    "--absolute",
    type=bool,
    is_flag=True,
    default=False,
    help="List absolute paths to log files (defaults to relative to the cwd).",
)
@click.option(
    "--recursive",
    type=bool,
    is_flag=True,
    default=True,
    help="List log files recursively (defaults to True).",
)
@click.option(
    "--json",
    type=bool,
    is_flag=True,
    default=False,
    help="Output listing as JSON",
)
@common_options
def logs(
    status: Literal["started", "success", "error"] | None,
    absolute: bool,
    recursive: bool,
    json: bool,
    **kwargs: Unpack[CommonOptions],
) -> None:
    """List log files in log directory."""
    (log_dir, log_level) = resolve_common_options(kwargs)

    # list the logs
    logs = list_eval_logs(
        log_dir=log_dir,
        filter=(lambda log: log.status == status) if status else None,
        recursive=recursive,
    )

    # convert file names
    for log in logs:
        if urlparse(log.name).scheme == "file":
            _, path = split_protocol(log.name)
            log.name = path
            if not absolute:
                log.name = os.path.relpath(log.name, os.path.curdir)

    if json:
        logs_dicts = [log.model_dump() for log in logs]
        print(dumps(logs_dicts, indent=2))

    else:
        for log in logs:
            print(log.name)



from pydantic_core import PydanticSerializationError
from utils import skip_if_no_openai

from inspect_ai import Task, eval
from inspect_ai.dataset import Sample
from inspect_ai.solver import (
    Generate,
    Plan,
    TaskState,
    generate,
    solver,
)


class NotSerializable:
    name: str


@skip_if_no_openai
def test_ignore_unserializable():
    @solver
    def inject_unserializable():
        async def solve(state: TaskState, generate: Generate):
            state.metadata["not serializable"] = NotSerializable
            return state

        return solve

    task = Task(
        dataset=[Sample(input="Say hello.", target="Hello")],
        plan=Plan(steps=[inject_unserializable(), generate()]),
    )

    try:
        eval(task, model="openai/gpt-4")
    except PydanticSerializationError:
        assert False, "Eval raised Pydantic serialization error."



import inspect
from importlib import import_module
from inspect import get_annotations, getmodule, isclass
from typing import Any, Callable, Literal, cast

from pydantic import BaseModel, Field

from .constants import PKG_NAME

RegistryType = Literal[
    "modelapi",
    "task",
    "solver",
    "plan",
    "scorer",
    "metric",
    "tool",
]


class RegistryInfo(BaseModel):
    type: RegistryType
    name: str
    metadata: dict[str, Any] = Field(default={})


def registry_add(o: object, info: RegistryInfo) -> None:
    r"""Add an object to the registry.

    Add the passed object to the registry using the RegistryInfo
    to index it for retreival. The RegistryInfo is also added
    to the object as an attribute, which can retrevied by calling
    registry_info() on an object instance.

    Args:
        o (object): Object to be registered (Metric, Solver, etc.)
        info (RegistryInfo): Metadata (name, etc.) for object.
    """
    # tag the object
    setattr(o, REGISTRY_INFO, info)

    # add to registry
    registry[registry_key(info.type, info.name)] = o


def registry_tag(
    type: Callable[..., Any],
    o: object,
    info: RegistryInfo,
    *args: list[Any],
    **kwargs: dict[str, Any],
) -> None:
    r"""Tag an object w/ registry info.

    Tag the passed object with RegistryInfo. This function DOES NOT
    add the object to the registry (call registry_add() to both
    tag and add an object to the registry). Call registry_info()
    on a tagged/registered object to retreive its info

    Args:
        type (T): type of object being tagged
        o (object): Object to be registered (Metric, Solver, etc.)
        info (RegistryInfo): Metadata (name, etc.) for object.
        *args (list[Any]): Creation arguments
        **kwargs (dict[str,Any]): Creation keyword arguments
    """
    # determine arg names and add them to kwargs
    named_params: dict[str, Any] = {}
    if len(args) > 0:
        params = list(inspect.signature(type).parameters.keys())
        for i, arg in enumerate(args):
            named_params[params[i]] = arg
    named_params |= kwargs

    # callables are not serializable so use their names
    for param in named_params.keys():
        if is_registry_object(named_params[param]):
            named_params[param] = registry_info(named_params[param]).name
        elif hasattr(named_params[param], "__name__"):
            named_params[param] = getattr(named_params[param], "__name__")
        else:
            named_params[param] = str(named_params[param])

    # set attribute
    setattr(o, REGISTRY_INFO, info)
    setattr(o, REGISTRY_PARAMS, named_params)


def registry_name(o: object, name: str) -> str:
    r"""Compute the registry name of an object.

    This function checks whether the passsed object is in a package,
    and if it is, preprends the package name as a namespace
    """
    package = get_package_name(o)
    return f"{package}/{name}" if package else name


def registry_lookup(type: RegistryType, name: str) -> object | None:
    r"""Lookup an object in the registry by type and name.

    Objects that defined in inspect extension packages (i.e. not
    directly within the core inspect_ai package) must be namespaced
    (e.g. "fancy_prompts/jailbreaker")

    Args:
        type: Type of object to find
        name: Name of object to find

    Returns:
        Object or None if not found.
    """
    # first try
    object = registry.get(registry_key(type, name))
    if object:
        return object
    # unnamespaced objects can also be found in inspect_ai
    elif name.find("/") == -1:
        return registry.get(registry_key(type, f"{PKG_NAME}/{name}"))
    else:
        return None


def registry_find(predicate: Callable[[RegistryInfo], bool]) -> list[object]:
    r"""Find objects in the registry that match the passed predicate.

    Args:
        predicate (Callable[[RegistryInfo], bool]): Predicate to find

    Returns:
        List of registry objects found
    """
    return [object for object in registry.values() if predicate(registry_info(object))]


def registry_create(type: RegistryType, name: str, **kwargs: Any) -> object:
    r"""Create a registry object.

    Registry objects can be ordinary functions that implement a protocol,
    factory functions that return a function based on **kwargs, or classes
    deriving that can be created using **kwargs

    Args:
        type (RegistryType): Type of registry object to create
        name (str): Name of registry options to create
        **kwargs (Any): Optional creation arguments

    Returns:
        Registry object with registry info attribute
    """
    # lookup the object
    obj = registry_lookup(type, name)

    # forward registry info to the instantiated object
    def with_registry_info(o: object) -> object:
        return set_registry_info(o, registry_info(obj))

    if isclass(obj):
        return with_registry_info(obj(**kwargs))
    elif callable(obj):
        return_type = getattr(get_annotations(obj)["return"], "__name__", None)
        if return_type and return_type.lower() == type:
            return with_registry_info(obj(**kwargs))
        else:
            return obj
    else:
        raise ValueError(f"{name} was not found in the registry")


def registry_info(o: object) -> RegistryInfo:
    r"""Lookup RegistryInfo for an object.

    Args:
        o (object): Object to lookup info for

    Returns:
        RegistryInfo for object.
    """
    info = getattr(o, REGISTRY_INFO)
    if info:
        return cast(RegistryInfo, info)
    else:
        raise ValueError("Object does not have registry info")


def registry_params(o: object) -> dict[str, Any]:
    r"""Lookup parameters used to instantiate a registry object.

    Args:
        o (object): Object to lookup info for

    Returns:
        Dictionary of parameters used to instantiate object.
    """
    params = getattr(o, REGISTRY_PARAMS)
    if params is not None:
        return cast(dict[str, Any], params)
    else:
        raise ValueError("Object does not have registry info")


def registry_log_name(o: object) -> str:
    r"""Name of object for logging.

    Registry objects defined by the inspect_ai package have their
    prefix stripped when written to the log (they in turn can also
    be created/referenced without the prefix).

    Args:
        o (object): Object to get name for

    Returns:
        Name of object for logging.
    """
    name = registry_info(o).name
    return name.replace(f"{PKG_NAME}/", "", 1)


def registry_unqualified_name(o: object) -> str:
    r"""Unqualfied name of object (i.e. without package prefix).

    Args:
        o (object): Object to get unqualfied name for

    Returns:
        Unqualfieid name of object
    """
    parts = registry_info(o).name.split("/")
    if len(parts) == 1:
        return parts[0]
    else:
        return "/".join(parts[1:])


def is_registry_object(o: object, type: RegistryType | None = None) -> bool:
    r"""Check if an object is a registry object.

    Args:
        o (object): Object to lookup info for
        type: (RegistryType | None): Optional. Check for a specific type

    Returns:
        True if the object is a registry object (optionally of the specified
        type). Otherwise, False
    """
    info = getattr(o, REGISTRY_INFO, None)
    if info:
        reg_info = cast(RegistryInfo, info)
        if type:
            return reg_info.type == type
        else:
            return True
    else:
        return False


def set_registry_info(o: object, info: RegistryInfo) -> object:
    r"""Set the RegistryInfo for an object.

    Args:
        o (object): Object to set the registry info for
        info: (object): Registry info

    Returns:
        Passed object, with RegistryInfo attached
    """
    setattr(o, REGISTRY_INFO, info)
    return o


def registry_key(type: RegistryType, name: str) -> str:
    return f"{type}:{name}"


REGISTRY_INFO = "__registry_info__"
REGISTRY_PARAMS = "__registry_params__"
registry: dict[str, object] = {}


def get_package_name(o: object) -> str | None:
    module = getmodule(o)
    package = str(getattr(module, "__package__", ""))
    if package:
        package = package.split(".")[0]
        if package != "None":
            package_module = import_module(package)
            if package_module:
                package_path = getattr(package_module, "__path__", None)
                if package_path:
                    return package

    return None



from typing import Callable, Literal

from inspect_ai._util.text import strip_numeric_punctuation, strip_punctuation
from inspect_ai.solver import TaskState

from ._metric import CORRECT, INCORRECT, Score
from ._scorer import Scorer, Target


def str_match_scorer(match: Callable[[str, str], tuple[str, bool]]) -> Scorer:
    """Scorer that uses a matching function.

    The matching function returns tuple[str,bool], where str is the answer
    extracted from the model output and bool is whether it matched the target
    """

    async def score(state: TaskState, target: Target) -> Score:
        answer: str | None = None
        for value in target:
            answer, matched = match(state.output.completion, value)
            if matched:
                return Score(
                    value=CORRECT, answer=answer, explanation=state.output.completion
                )

        return Score(
            value=INCORRECT, answer=answer, explanation=state.output.completion
        )

    return score


def match_str(
    value: str,
    target: str,
    location: Literal["begin", "end", "any", "exact"] = "end",
    ignore_case: bool = True,
    ignore_punctuation: bool = True,
    numeric: bool = False,
) -> tuple[str, bool]:
    # strip ws
    v = value.strip()
    t = target.strip()

    # baseline answer (will only change for numeric)
    answer = v

    # further cleanup
    if ignore_case:
        v = v.lower()
        t = t.lower()
    if numeric:
        # remove punctuation
        v = strip_numeric_punctuation(v)
        t = strip_numeric_punctuation(t)
        # normalize as required
        t = normalize_number(t)
        if location == "begin":
            words = v.split(" ")
            v = first_number_normalized(words)
        elif location == "end":
            words = v.split(" ")
            words.reverse()
            v = first_number_normalized(words)
        elif location == "exact":
            v = normalize_number(v)
        answer = v
    elif ignore_punctuation:
        v = strip_punctuation(v)
        t = strip_punctuation(t)

    # comparisons
    if location == "begin":
        return answer, v.startswith(t)
    elif location == "end":
        return answer, v.endswith(t)
    elif location == "exact":
        return answer, v == t
    else:
        return answer, t in v


def first_number_normalized(words: list[str]) -> str:
    number = next((word for word in words if word.isnumeric()), words[0])
    return normalize_number(number)


def normalize_number(number: str, precision: int = 5) -> str:
    if number.replace(".", "").isnumeric():
        num = float(number)
        return format(num, f".{precision}g")
    else:
        return number



import inspect
from typing import Any, Awaitable, Callable, TypeVar, cast

from inspect_ai._util.registry import (
    RegistryInfo,
    is_registry_object,
    registry_add,
    registry_create,
    registry_info,
    registry_name,
    registry_tag,
)

from ._solver import Solver, TaskState


class Plan:
    """Task plan: List of solvers with an optional finishing solver.

    The optional `finish` solver is called after executing the steps (including in the case
    where the steps were exited early due to `TaskState.completed = True` or `max_messages`).

    The optional `cleanup` function is called when the plan is complete (even if the plan
    is terminated due to an exception).
    """

    def __init__(
        self,
        steps: Solver | list[Solver],
        finish: Solver | None = None,
        cleanup: Callable[[TaskState], Awaitable[None]] | None = None,
        name: str | None = None,
    ) -> None:
        """Create a task plan.

        Args:
          steps (list[Solver]): Solvers to run for this plan.
          finish (Solver | None): Finishing solver that is always run even for early exit.
            Note that this solver is NOT run when exception are thrown (use `cleanup` for this)
          cleanup (Callable[[TaskState], Awaitable[None]] | None): Optional cleanup handler that
            is called at the end (even if an exception occurs). Note that this function takes
            a `TaskState` but does not return one (it is only for cleanup not for transforming
            the state).
          name (str | None): Optional name for plan (for log files).
        """
        if isinstance(steps, Solver):
            self.steps = [steps]
        else:
            self.steps = steps

        self.finish = finish
        self.cleanup = cleanup
        self._name = name

    @property
    def name(self) -> str:
        if self._name is not None:
            return self._name
        elif is_registry_object(self):
            return registry_info(self).name
        else:
            return "plan"

    steps: list[Solver]
    """Solvers to run for this plan."""

    finish: Solver | None = None
    """Finishing sover that is always run even for early exit."""

    cleanup: Callable[[TaskState], Awaitable[None]] | None = None
    """Function  called at the end of the plan (even if an exception occurs).

    Note that this function takes a `TaskState` but does not return one
    (it is only for cleanup not for transforming the state). Note also that
    this function should be declared `async`.
    """


PlanType = TypeVar("PlanType", bound=Callable[..., Plan])


def plan(*plan: PlanType | None, name: str | None = None, **attribs: Any) -> Any:
    r"""Decorator for registering plans.

    Args:
      *plan (PlanType): Function returning `Plan` targeted by
        plain plan decorator without attributes (e.g. `@plan`)
      name (str | None):
        Optional name for plan. If the decorator has no name
        argument then the name of the function
        will be used to automatically assign a name.
      **attribs: (dict[str,Any]): Additional plan attributes.

    Returns:
        Plan with registry attributes.
    """

    def create_plan_wrapper(plan_type: PlanType) -> PlanType:
        # get the name and params
        plan_name = registry_name(plan_type, name or getattr(plan_type, "__name__"))
        params = list(inspect.signature(plan_type).parameters.keys())

        # create and return the wrapper
        def wrapper(*w_args: Any, **w_kwargs: Any) -> Plan:
            # create the plan
            plan = plan_type(*w_args, **w_kwargs)

            # tag it
            registry_tag(
                plan_type,
                plan,
                RegistryInfo(
                    type="plan",
                    name=plan_name,
                    metadata=dict(attribs=attribs, params=params),
                ),
                *w_args,
                **w_kwargs,
            )

            # return it
            return plan

        return plan_register(
            plan=cast(PlanType, wrapper), name=plan_name, attribs=attribs, params=params
        )

    if plan:
        return create_plan_wrapper(cast(PlanType, plan[0]))
    else:
        return create_plan_wrapper


def plan_register(
    plan: PlanType, name: str, attribs: dict[str, Any], params: list[str]
) -> PlanType:
    r"""Register a plan.

    Args:
        plan (PlanType): function that returns a Plan
        name (str): Name of plan
        attribs (dict[str,Any]): Attributes of plan decorator
        params (list[str]): Plan parameter names

    Returns:
        Plan with registry attributes.
    """
    registry_add(
        plan,
        RegistryInfo(
            type="plan", name=name, metadata=dict(attribs=attribs, params=params)
        ),
    )
    return plan


def plan_create(name: str, **kwargs: Any) -> Plan:
    r"""Create a Plan based on its registered name.

    Args:
        name (str): Name of plan
        **kwargs (dict): Optional creation arguments for the plan

    Returns:
        Plan with registry info attribute
    """
    return cast(Plan, registry_create("plan", name, **kwargs))



import errno
from typing import Any, Literal
from urllib.parse import urlparse
from urllib.request import url2pathname

from inspect_ai._util.file import file, filesystem


def resource(
    resource: str,
    type: Literal["auto", "file"] = "auto",
    fs_options: dict[str, Any] = {},
) -> str:
    """Read and resolve a resource to a string.

    Resources are often used for templates, configuration, etc.
    They are sometimes hard-coded strings, and sometimes paths
    to external resources (e.g. in the local filesystem or
    remote stores e.g. s3:// or https://).

    The `resource()` function will resolve its argument to
    a resource string. If a protocol-prefixed file name
    (e.g. s3://) or the path to a local file that exists
    is passed then it will be read and its contents returned.
    Otherwise, it will return the passed `str` directly
    This function is mostly intended as a helper for other
    functions that take either a string or a resource path
    as an argument, and want to easily resolve them to
    the underlying content.

    If you want to ensure that only local or remote files
    are consumed, specify `type="file"`. For example:
    `resource("templates/prompt.txt", type="file")`

    Args:
        resource (str): Path to local or remote (e.g. s3://)
          resource, or for `type="auto"` (the default),
          a string containing the literal resource value.
        type (Literal["auto", "file"]): For "auto" (the default),
          interpret the resource as a literal string if its not
          a valid path. For "file", always interpret it as
          a file path.
        fs_options (dict[str, Any]): Optional. Addional
          arguments to pass through to the `fsspec` filesystem
          provider (e.g. `S3FileSystem`). Use `{"anon": True }`
          if you are accessing a public S3 bucket with no
          credentials.

    Returns:
       Text content of resource.
    """

    # helper function to read the resource as a file
    def read_resource() -> str:
        with file(resource, "r", fs_options=fs_options) as f:
            return f.read()

    if type == "file":
        return read_resource()
    else:
        # parse the url
        try:
            parsed = urlparse(resource)
        except OSError:
            return resource

        # if it has a scheme then its likely a file
        if parsed.scheme:
            try:
                return read_resource()
            except FileNotFoundError:
                return resource
            except OSError as ex:
                if ex.errno == errno.ENAMETOOLONG:
                    return resource
                else:
                    raise ex

        # no scheme means either a local file or a string
        else:
            # extract the path
            try:
                path = url2pathname(parsed.path)
            except OSError:
                return resource

            # return it if it exists (otherwise return the str)
            fs = filesystem(path)
            if fs.exists(path):
                return read_resource()
            else:
                return resource



import asyncio
import os
import shlex
import sys
from contextvars import ContextVar
from dataclasses import dataclass
from pathlib import Path
from typing import Generic, Literal, TypeVar, Union, overload

from .concurrency import concurrency, using_concurrency

T = TypeVar("T", str, bytes)


@dataclass
class ProcessResult(Generic[T]):
    success: bool
    """Did the process exit with success."""

    returncode: int
    """Return code from process exit."""

    stdout: T
    """Contents of stdout."""

    stderr: T
    """Contents of stderr."""


@overload
# type: ignore
async def subprocess(
    args: str | list[str],
    text: Literal[True] = True,
    input: str | bytes | memoryview | None = None,
    cwd: str | Path | None = None,
    env: dict[str, str] = {},
    timeout: int | None = None,
) -> ProcessResult[str]:
    ...


@overload
async def subprocess(
    args: str | list[str],
    text: Literal[False] = False,
    input: str | bytes | memoryview | None = None,
    cwd: str | Path | None = None,
    env: dict[str, str] = {},
    timeout: int | None = None,
) -> ProcessResult[bytes]:
    ...


async def subprocess(
    args: str | list[str],
    text: bool = True,
    input: str | bytes | memoryview | None = None,
    cwd: str | Path | None = None,
    env: dict[str, str] = {},
    timeout: int | None = None,
) -> Union[ProcessResult[str], ProcessResult[bytes]]:
    """Execute and wait for a subprocess.

    Convenience method for solvers, scorers, and tools to launch
    subprocesses. Automatically enforces a limit on concurrent
    subprocesses (defaulting to os.cpu_count() but controllable
    via the `max_subproccesses` eval config option).

    Args:
       args (str | list[str]): Command and arguments to execute.
       text (bool): Return stdout and stderr as text (defaults to True)
       input (str | bytes | memoryview | None): Optional stdin
          for subprocess.
       cwd (str | Path | None): Switch to directory for execution.
       env (dict[str, str]): Additional environment variables.
       timeout (int | None): Timeout

    Returns:
       Subprocess result (text or binary depending on `text` param)
    """
    # resolve input
    input = input.encode() if isinstance(input, str) else input

    # build command
    args = args if isinstance(args, list) else [args]
    command = " ".join([shlex.quote(arg) for arg in args])

    # function to run command (we may or may not run it w/ concurrency)
    async def run_command() -> Union[ProcessResult[str], ProcessResult[bytes]]:
        proc = await asyncio.create_subprocess_shell(
            command,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=cwd,
            env={**os.environ, **env},
        )

        # wait for it to execute and return result
        stdout, stderr = await proc.communicate(input=input)
        success = proc.returncode == 0
        returncode = proc.returncode if proc.returncode is not None else 1
        if text:
            return ProcessResult[str](
                success=success,
                returncode=returncode,
                stdout=stdout.decode(),
                stderr=stderr.decode(),
            )
        else:
            return ProcessResult[bytes](
                success=success,
                returncode=returncode,
                stdout=stdout,
                stderr=stderr,
            )

    # wrapper for run command that implements timeout
    async def run_command_timeout() -> Union[ProcessResult[str], ProcessResult[bytes]]:
        if timeout:
            try:
                if sys.version_info >= (3, 11):
                    async with asyncio.timeout(timeout):
                        return await run_command()
                else:
                    return await asyncio.wait_for(run_command(), timeout=timeout)
            except asyncio.exceptions.TimeoutError:
                return ProcessResult(
                    False, 1, "", "Command timed out before completing"
                )
        else:
            return await run_command()

    # run command
    if using_concurrency():
        async with concurrency("subprocesses", max_subprocesses_context_var.get()):
            return await run_command_timeout()
    else:
        return await run_command_timeout()


def init_subprocess(max_subprocesses: int | None = None) -> None:
    # initialize dedicated subprocesses semaphore
    cpus = os.cpu_count()
    max_subprocesses = max_subprocesses if max_subprocesses else cpus if cpus else 1
    max_subprocesses_context_var.set(max_subprocesses)


max_subprocesses_context_var = ContextVar[int]("max_subprocesses")



"""
Measuring Massive Multitask Language Understanding

Dan Hendrycks, Collin Burns, Steven Basart, Andy Zou,
Mantas Mazeika, Dawn Song, Jacob Steinhardt
https://arxiv.org/abs/2009.03300

Based on: https://github.com/openai/simple-evals/blob/main/mmlu_eval.py

# eval all subjects w/ 500 randomly selected samples
inspect eval mmlu.py@mmlu --limit 500

# add chain of thought
inspect eval mmlu.py@mmlu --limit 500 -T cot=true

# eval selected subjects
inspect eval mmlu.py@mmlu -T subjects=anatomy,astronomy

# eval single subjects
inspect eval mmlu.py@mmlu_anatomy
inspect eval mmlu.py@mmlu_astronomy
"""

from inspect_ai import Task, task
from inspect_ai.dataset import Sample, csv_dataset
from inspect_ai.model import GenerateConfig
from inspect_ai.scorer import answer
from inspect_ai.solver import multiple_choice


# map records to inspect sample
def record_to_sample(record):
    return Sample(
        input=record["Question"],
        choices=[
            str(record["A"]),
            str(record["B"]),
            str(record["C"]),
            str(record["D"]),
        ],
        target=record["Answer"],
        metadata={"subject": record["Subject"]},
    )


# read dataset globally so it can be shared by all of the tasks
# (shuffle so that --limit draws from multiple subjects)
dataset = csv_dataset(
    csv_file="datasets/mmlu.csv",
    sample_fields=record_to_sample,
    shuffle=True,
)


@task
def mmlu(subjects=[], cot=False):
    # filter dataset if requested
    subjects = subjects if isinstance(subjects, list) else [subjects]
    if len(subjects) > 0:
        task_dataset = dataset.filter(
            name=f"{dataset.name}-{'-'.join(subjects)}",
            predicate=lambda sample: sample.metadata["subject"] in subjects,
        )
    else:
        task_dataset = dataset

    # return task
    return Task(
        dataset=task_dataset,
        plan=multiple_choice(cot=cot),
        scorer=answer("letter"),
        config=GenerateConfig(temperature=0.5),
    )


@task
def mmlu_abstract_algebra(cot=False):
    return mmlu("abstract_algebra", cot)


@task
def mmlu_anatomy(cot=False):
    return mmlu("anatomy", cot)


@task
def mmlu_astronomy(cot=False):
    return mmlu("astronomy", cot)


@task
def mmlu_business_ethics(cot=False):
    return mmlu("business_ethics", cot)


@task
def mmlu_clinical_knowledge(cot=False):
    return mmlu("clinical_knowledge", cot)


@task
def mmlu_college_biology(cot=False):
    return mmlu("college_biology", cot)


@task
def mmlu_college_chemistry(cot=False):
    return mmlu("college_chemistry", cot)


@task
def mmlu_college_computer_science(cot=False):
    return mmlu("college_computer_science", cot)


@task
def mmlu_college_mathematics(cot=False):
    return mmlu("college_mathematics", cot)


@task
def mmlu_college_medicine(cot=False):
    return mmlu("college_medicine", cot)


@task
def mmlu_college_physics(cot=False):
    return mmlu("college_physics", cot)


@task
def mmlu_computer_security(cot=False):
    return mmlu("computer_security", cot)


@task
def mmlu_conceptual_physics(cot=False):
    return mmlu("conceptual_physics", cot)


@task
def mmlu_electrical_engineering(cot=False):
    return mmlu("electrical_engineering", cot)


@task
def mmlu_elementary_mathematics(cot=False):
    return mmlu("elementary_mathematics", cot)


@task
def mmlu_formal_logic(cot=False):
    return mmlu("formal_logic", cot)


@task
def mmlu_global_facts(cot=False):
    return mmlu("global_facts", cot)


@task
def mmlu_high_school_biology(cot=False):
    return mmlu("high_school_biology", cot)


@task
def mmlu_high_school_chemistry(cot=False):
    return mmlu("high_school_chemistry", cot)


@task
def mmlu_high_school_computer_science(cot=False):
    return mmlu("high_school_computer_science", cot)


@task
def mmlu_high_school_european_history(cot=False):
    return mmlu("high_school_european_history", cot)


@task
def mmlu_high_school_geography(cot=False):
    return mmlu("high_school_geography", cot)


@task
def mmlu_high_school_government_and_politics(cot=False):
    return mmlu("high_school_government_and_politics", cot)


@task
def mmlu_high_school_macroeconomics(cot=False):
    return mmlu("high_school_macroeconomics", cot)


@task
def mmlu_high_school_mathematics(cot=False):
    return mmlu("high_school_mathematics", cot)


@task
def mmlu_high_school_microeconomics(cot=False):
    return mmlu("high_school_microeconomics", cot)


@task
def mmlu_high_school_physics(cot=False):
    return mmlu("high_school_physics", cot)


@task
def mmlu_high_school_psychology(cot=False):
    return mmlu("high_school_psychology", cot)


@task
def mmlu_high_school_statistics(cot=False):
    return mmlu("high_school_statistics", cot)


@task
def mmlu_high_school_us_history(cot=False):
    return mmlu("high_school_us_history", cot)


@task
def mmlu_high_school_world_history(cot=False):
    return mmlu("high_school_world_history", cot)


@task
def mmlu_human_aging(cot=False):
    return mmlu("human_aging", cot)


@task
def mmlu_human_sexuality(cot=False):
    return mmlu("human_sexuality", cot)


@task
def mmlu_international_law(cot=False):
    return mmlu("international_law", cot)


@task
def mmlu_jurisprudence(cot=False):
    return mmlu("jurisprudence", cot)


@task
def mmlu_logical_fallacies(cot=False):
    return mmlu("logical_fallacies", cot)


@task
def mmlu_machine_learning(cot=False):
    return mmlu("machine_learning", cot)


@task
def mmlu_management(cot=False):
    return mmlu("management", cot)


@task
def mmlu_marketing(cot=False):
    return mmlu("marketing", cot)


@task
def mmlu_miscellaneous(cot=False):
    return mmlu("miscellaneous", cot)


@task
def mmlu_moral_disputes(cot=False):
    return mmlu("moral_disputes", cot)


@task
def mmlu_moral_scenarios(cot=False):
    return mmlu("moral_scenarios", cot)


@task
def mmlu_nutrition(cot=False):
    return mmlu("nutrition", cot)


@task
def mmlu_philosophy(cot=False):
    return mmlu("philosophy", cot)


@task
def mmlu_prehistory(cot=False):
    return mmlu("prehistory", cot)


@task
def mmlu_professional_accounting(cot=False):
    return mmlu("professional_accounting", cot)


@task
def mmlu_professional_law(cot=False):
    return mmlu("professional_law", cot)


@task
def mmlu_professional_medicine(cot=False):
    return mmlu("professional_medicine", cot)


@task
def mmlu_professional_psychology(cot=False):
    return mmlu("professional_psychology", cot)


@task
def mmlu_public_relations(cot=False):
    return mmlu("public_relations", cot)


@task
def mmlu_security_studies(cot=False):
    return mmlu("security_studies", cot)


@task
def mmlu_sociology(cot=False):
    return mmlu("sociology", cot)


@task
def mmlu_us_foreign_policy(cot=False):
    return mmlu("us_foreign_policy", cot)


@task
def mmlu_virology(cot=False):
    return mmlu("virology", cot)


@task
def mmlu_world_religions(cot=False):
    return mmlu("world_religions", cot)



import ast
import builtins
import os
import re
from copy import deepcopy
from typing import Any, Tuple, cast
from xml.sax.saxutils import escape

from anthropic import (
    APIConnectionError,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    BadRequestError,
    InternalServerError,
    RateLimitError,
)
from anthropic._types import NOT_GIVEN
from anthropic.types import (
    ImageBlockParam,
    Message,
    MessageParam,
    TextBlock,
    TextBlockParam,
)
from anthropic.types.beta.tools import ToolParam as BetaToolParam
from anthropic.types.beta.tools import (
    ToolResultBlockParam,
    ToolsBetaMessage,
    ToolsBetaMessageParam,
    ToolUseBlock,
    ToolUseBlockParam,
)
from anthropic.types.beta.tools.tool_param import (
    InputSchema,
)
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_RETRIES, DEFAULT_MAX_TOKENS
from inspect_ai._util.error import exception_message
from inspect_ai._util.images import image_as_data_uri
from inspect_ai._util.json import json_type_to_python_type
from inspect_ai._util.url import data_uri_mime_type, data_uri_to_base64, is_data_uri
from inspect_ai.model._providers.util import model_base_url

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
    Content,
    ContentText,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
    ModelUsage,
    StopReason,
)
from .._tool import ToolCall, ToolChoice, ToolFunction, ToolInfo, ToolParam
from .._util import chat_api_tool

ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"


class AnthropicAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None,
        config: GenerateConfig = GenerateConfig(),
        bedrock: bool = False,
        tools_beta: bool = True,
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        self.tools_beta = tools_beta and not bedrock

        # create client
        if bedrock:
            base_url = model_base_url(
                base_url, ["ANTHROPIC_BEDROCK_BASE_URL", "BEDROCK_ANTHROPIC_BASE_URL"]
            )

            self.client: AsyncAnthropic | AsyncAnthropicBedrock = AsyncAnthropicBedrock(
                base_url=base_url,
                max_retries=(
                    config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES
                ),
                **model_args,
            )
        else:
            # resolve api_key
            api_key = os.environ.get(ANTHROPIC_API_KEY, None)
            if api_key is None:
                raise ValueError(f"{ANTHROPIC_API_KEY} environment variable not found.")
            self.api_key = api_key
            base_url = model_base_url(base_url, "ANTHROPIC_BASE_URL")
            self.client = AsyncAnthropic(
                base_url=base_url,
                api_key=self.api_key,
                max_retries=(
                    config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES
                ),
                **model_args,
            )

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # generate
        try:
            # use tools beta endpoint if we have tools and haven't opted out (note that
            # bedrock is an implicit opt-out as it doesn't yet support the tools api
            if (
                len(tools) > 0
                and self.tools_beta
                and not isinstance(self.client, AsyncAnthropicBedrock)
            ):
                (
                    system_message,
                    beta_tools,
                    beta_messages,
                ) = await resolve_tools_beta_chat_input(
                    input, tools, tool_choice, config
                )

                message = await self.client.beta.tools.messages.create(
                    stream=False,
                    messages=beta_messages,
                    system=system_message if system_message is not None else NOT_GIVEN,
                    stop_sequences=(
                        config.stop_seqs if config.stop_seqs is not None else NOT_GIVEN
                    ),
                    tools=beta_tools,
                    **self.completion_params(config),
                )

                return tools_beta_model_output_from_message(message, tools)

            # otherwise use standard chat endpoint
            else:
                system_message, stop_seq, messages = await resolve_chat_input(
                    input, tools, config
                )

                message = await self.client.messages.create(
                    stream=False,
                    messages=messages,
                    system=system_message if system_message is not None else NOT_GIVEN,
                    stop_sequences=stop_seq if stop_seq is not None else NOT_GIVEN,
                    **self.completion_params(config),
                )

                # extract model output from text response (may have tool calls)
                return model_output_from_message(message, tools)

        except BadRequestError as ex:
            return ModelOutput.from_content(
                model=self.model_name,
                content="Sorry, but I can't assist with that",
                stop_reason="content_filter",
                error=exception_message(ex),
            )

    def completion_params(self, config: GenerateConfig) -> dict[str, Any]:
        return dict(
            model=self.model_name,
            max_tokens=cast(int, config.max_tokens),
            temperature=(
                config.temperature if config.temperature is not None else NOT_GIVEN
            ),
            top_p=config.top_p if config.top_p is not None else NOT_GIVEN,
            top_k=config.top_k if config.top_k is not None else NOT_GIVEN,
            timeout=float(config.timeout) if config.timeout is not None else NOT_GIVEN,
        )

    @override
    def max_tokens(self) -> int | None:
        # anthropic requires you to expicitly specify max_tokens (most others
        # set it to the maximum allowable output tokens for the model).
        return DEFAULT_MAX_TOKENS

    @override
    def connection_key(self) -> str:
        return self.api_key

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        # We have observed that anthropic will frequently return InternalServerError
        # seeminly in place of RateLimitError (at the very least the errors seem to
        # always be transient). Equating this to rate limit errors may occationally
        # result in retrying too many times, but much more often will avert a failed
        # eval that just needed to survive a transient error
        return (
            isinstance(ex, RateLimitError)
            or isinstance(ex, InternalServerError)
            or isinstance(ex, APIConnectionError)
        )

    @override
    def collapse_user_messages(self) -> bool:
        return True


#######################################################################################
# Resolve input, tools, and config into the right shape of input for the Anthropic
# tool use beta. we also keep the legacy tools implementation around for now (see below)
# for users on Bedrock of who want to opt out for tools beta for any reason
#######################################################################################


async def resolve_tools_beta_chat_input(
    input: list[ChatMessage],
    tools: list[ToolInfo],
    tool_choice: ToolChoice,
    config: GenerateConfig,
) -> Tuple[str | None, list[BetaToolParam], list[ToolsBetaMessageParam]]:
    # extract system message
    system_message, messages = split_system_message(input, config)

    # some special handling for tools
    if len(tools) > 0:
        # encourage claude to show its thinking, see
        # https://docs.anthropic.com/claude/docs/tool-use#chain-of-thought-tool-use
        system_message = f"{system_message}\n\nBefore answering, explain your reasoning step-by-step."

        # implement tool_choice by appending to the last user message, see
        # https://docs.anthropic.com/claude/docs/tool-use#forcing-tool-use
        if isinstance(tool_choice, ToolFunction):
            messages = deepcopy(messages)
            message = next(
                (
                    message
                    for message in reversed(messages)
                    if isinstance(message, ChatMessageUser)
                ),
                None,
            )
            if message:
                message.text = (
                    f"{message.text} Use the {tool_choice.name} tool in your response."
                )

    # messages
    beta_messages = [(await tools_beta_message_param(message)) for message in messages]

    # tools
    chat_functions = [chat_api_tool(tool)["function"] for tool in tools]
    beta_tools = [
        BetaToolParam(
            name=function["name"],
            description=function["description"],
            input_schema=cast(InputSchema, function["parameters"]),
        )
        for function in chat_functions
    ]

    return system_message, beta_tools, beta_messages


async def tools_beta_message_param(message: ChatMessage) -> ToolsBetaMessageParam:
    # no system role for anthropic (this is more like an asseration,
    # as these should have already been filtered out)
    if message.role == "system":
        raise ValueError("Antropic models do not support the system role")

    # "tool" means serving a tool call result back to claude
    elif message.role == "tool":
        if message.tool_error is not None:
            content: str | list[TextBlockParam] = message.tool_error
        if isinstance(message.content, str):
            content = [TextBlockParam(type="text", text=message.content)]
        else:
            content = [
                TextBlockParam(type="text", text=content.text)
                for content in message.content
                if isinstance(content, ContentText)
            ]

        return ToolsBetaMessageParam(
            role="user",
            content=[
                ToolResultBlockParam(
                    tool_use_id=str(message.tool_call_id),
                    type="tool_result",
                    content=content,
                    is_error=message.tool_error is not None,
                )
            ],
        )

    # tool_calls means claude is attempting to call our tools
    elif message.role == "assistant" and message.tool_calls:
        # first include content (claude <thinking>)
        tools_content: list[TextBlockParam | ImageBlockParam | ToolUseBlockParam] = (
            [TextBlockParam(type="text", text=message.content)]
            if isinstance(message.content, str)
            else (
                [(await message_param_content(content)) for content in message.content]
            )
        )

        # now add tools
        for tool_call in message.tool_calls:
            tools_content.append(
                ToolUseBlockParam(
                    type="tool_use",
                    id=tool_call.id,
                    name=tool_call.function,
                    input=tool_call.arguments,
                )
            )

        return ToolsBetaMessageParam(
            role=message.role,
            content=tools_content,
        )

    # normal text content
    elif isinstance(message.content, str):
        return ToolsBetaMessageParam(role=message.role, content=message.content)

    # mixed text/images
    else:
        return ToolsBetaMessageParam(
            role=message.role,
            content=[
                await message_param_content(content) for content in message.content
            ],
        )


def tools_beta_model_output_from_message(
    message: ToolsBetaMessage, tools: list[ToolInfo]
) -> ModelOutput:
    # extract content and tool calls
    content: list[Content] = []
    tool_calls: list[ToolCall] | None = None

    for content_block in message.content:
        if isinstance(content_block, TextBlock):
            # if this was a tool call then remove <result></result> tags that
            # claude sometimes likes to insert!
            content_text = content_block.text
            if len(tools) > 0:
                content_text = content_text.replace("<result>", "").replace(
                    "</result>", ""
                )
            content.append(ContentText(type="text", text=content_text))
        elif isinstance(content_block, ToolUseBlock):
            tool_calls = tool_calls or []
            tool_calls.append(
                ToolCall(
                    type="function",
                    id=content_block.id,
                    function=content_block.name,
                    arguments=content_block.model_dump().get("input", {}),
                )
            )

    # resolve choice
    choice = ChatCompletionChoice(
        message=ChatMessageAssistant(
            content=content, tool_calls=tool_calls, source="generate"
        ),
        stop_reason=tools_beta_message_stop_reason(message),
    )

    # return ModelOutput
    return ModelOutput(
        model=message.model,
        choices=[choice],
        usage=ModelUsage(
            input_tokens=message.usage.input_tokens,
            output_tokens=message.usage.output_tokens,
            total_tokens=message.usage.input_tokens + message.usage.output_tokens,
        ),
    )


def tools_beta_message_stop_reason(message: ToolsBetaMessage) -> StopReason:
    match message.stop_reason:
        case "end_turn" | "stop_sequence":
            return "stop"
        case "max_tokens":
            return "length"
        case "tool_use":
            return "tool_calls"
        case _:
            return "unknown"


def split_system_message(
    input: list[ChatMessage], config: GenerateConfig
) -> Tuple[str | None, list[ChatMessage]]:
    # split messages
    system_messages = [m for m in input if isinstance(m, ChatMessageSystem)]
    messages = [m for m in input if not isinstance(m, ChatMessageSystem)]

    # build system message
    system_message = (
        "\n\n".join([message.text for message in system_messages])
        if len(system_messages) > 0
        else None
    )

    # prepend any config based system message
    if config.system_message:
        system_message = f"{config.system_message}\n\n{system_message}"

    # return
    return system_message, cast(list[ChatMessage], messages)


#######################################################################################
# Resolve input, tools, and config into the right shape of input for Anthropic models.
#
# Anthropic tools are defined not using a tools component of their API, but rather by
# defineing all available tools in the system message. If there are tools then there
# is also a requirement to define a custom stop sequence. This fucntion sorts all of
# that out and returns a system message, a stop sequence (if necessary) and the list
# of anthropic-native MessageParam objects (including converting role="tool" messages
# into XML encoded role="user" messages for Claude
#######################################################################################

FUNCTIONS_STOP_SEQ = "</function_calls>"


async def resolve_chat_input(
    input: list[ChatMessage], tools: list[ToolInfo], config: GenerateConfig
) -> Tuple[str | None, list[str] | None, list[MessageParam]]:
    # extract system message
    system_message, messages = split_system_message(input, config)

    # resolve tool use (system message and stop sequences)
    stop_seqs = deepcopy(config.stop_seqs)
    if len(tools) > 0:
        system_message = f"{system_message}\n\n{tools_system_message(tools)}"
        stop_seqs = (
            config.stop_seqs if config.stop_seqs else ["\n\nHuman:", "\n\nAssistant"]
        )
        stop_seqs.append(FUNCTIONS_STOP_SEQ)

    # create anthropic message params
    message_params = [await message_param(m) for m in messages]

    # done!
    return system_message, stop_seqs, message_params


def tools_system_message(tools: list[ToolInfo]) -> str:
    tool_sep = "\n\n"
    return f"""
In this environment you have access to a set of tools you can use to answer the user's question.

You may call them like this:
<function_calls>
<invoke>
<tool_name>$TOOL_NAME</tool_name>
<parameters>
<$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>
...
</parameters>
</invoke>
</function_calls>

Here are the tools available:
<tools>
{tool_sep.join([tool_description(tool) for tool in tools])}
</tools>
"""


def tool_description(tool: ToolInfo) -> str:
    newline = "\n"
    return f"""
<tool_description>
<tool_name>{escape(tool.name)}</tool_name>
<description>{escape(tool.description)}</description>
<parameters>
{newline.join(tool_param(param) for param in tool.params)}
</parameter>
</tool_description>
"""


def tool_param(param: ToolParam) -> str:
    return f"""
<parameter>
<name>{escape(param.name)}</name>
<type>{escape(param.type)}</type>
<description>{escape(param.description)}</description>
</parameter>
"""


async def message_param(message: ChatMessage) -> MessageParam:
    # no system role for anthropic (this is more like an assertion,
    # as these should have already been filtered out)
    if message.role == "system":
        raise ValueError("Antropic models do not support the system role")

    # "tool" means serving a tool call result back to claude
    elif message.role == "tool":
        return tool_message_param(message)

    # tool_calls means claude is attempting to call our tools
    elif message.role == "assistant" and message.tool_calls:
        return MessageParam(
            role=message.role,
            content=f"{message.content}\n{function_calls(message.tool_calls)}",
        )

    # normal text content
    elif isinstance(message.content, str):
        return MessageParam(role=message.role, content=message.content)

    # mixed text/images
    else:
        return MessageParam(
            role=message.role,
            content=[
                await message_param_content(content) for content in message.content
            ],
        )


async def message_param_content(
    content: Content,
) -> TextBlockParam | ImageBlockParam:
    if isinstance(content, ContentText):
        return TextBlockParam(type="text", text=content.text)
    else:
        # resolve to url
        image = content.image
        if not is_data_uri(image):
            image = await image_as_data_uri(image)

        # resolve mime type and base64 content
        media_type = data_uri_mime_type(image) or "image/png"
        image = data_uri_to_base64(image)

        if media_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
            raise ValueError(f"Unable to read image of type {media_type}")

        return ImageBlockParam(
            type="image",
            source=dict(type="base64", media_type=cast(Any, media_type), data=image),
        )


def tool_message_param(message: ChatMessageTool) -> MessageParam:
    results = f"""
<function_results>
{function_result(message)}
</function_results>
"""
    return MessageParam(role="user", content=results)


def function_calls(tool_calls: list[ToolCall]) -> str:
    nl = "\n"
    return f"""
<function_calls>
{nl.join([function_call(tool_call) for tool_call in tool_calls])}
</function_calls>
"""


def function_call(tool_call: ToolCall) -> str:
    nl = "\n"
    return f"""
<invoke>
<tool_name>{escape(tool_call.function)}</tool_name>
<parameters>
{nl.join([function_parameter(name,value) for name, value in tool_call.arguments.items()])}
</parameters>
</invoke>
"""


def function_parameter(name: str, value: Any) -> str:
    return f"<{name}>{value}</{name}>"


def function_result(message: ChatMessageTool) -> str:
    if message.tool_error:
        return f"""
<system>
{escape(message.tool_error)}
</system>
"""
    else:
        return f"""
<result>
<tool_name>{escape(str(message.tool_call_id))}</tool_name>
<stdout>
{escape(message.text)}
</stdout>
</result>
"""


#######################################################################################
# Extract model output (including tool calls) from an Anthropic message
#
# Anthropic encodes tool calls (in XML) directly in role="assistant" messages. The
# code below deals with this by parsing out the tool calls and separating them into
# the Inspect native ToolCall objects.
#######################################################################################


def model_output_from_message(message: Message, tools: list[ToolInfo]) -> ModelOutput:
    # extract function calls (if any); throws ValueError if xml is invalid
    try:
        content_with_functions = extract_function_calls(message)
        if content_with_functions:
            content = content_with_functions.content
            tool_calls = [
                tool_call(function_call, tools)
                for function_call in content_with_functions.function_calls
            ]
        else:
            content = message_content(message)
            tool_calls = None
    except ValueError as ex:
        return ModelOutput.from_content(
            message.model,
            f"{message_content(message)}\n\nError: {exception_message(ex)}",
        )

    # resolve choice
    choice = ChatCompletionChoice(
        message=ChatMessageAssistant(
            content=content, tool_calls=tool_calls, source="generate"
        ),
        stop_reason=message_stop_reason(message),
    )

    # return ModelOutput
    return ModelOutput(
        model=message.model,
        choices=[choice],
        usage=ModelUsage(
            input_tokens=message.usage.input_tokens,
            output_tokens=message.usage.output_tokens,
            total_tokens=message.usage.input_tokens + message.usage.output_tokens,
        ),
    )


def message_stop_reason(message: Message) -> StopReason:
    match message.stop_reason:
        case "end_turn":
            return "stop"
        case "max_tokens":
            return "length"
        case "stop_sequence":
            if message.stop_sequence == FUNCTIONS_STOP_SEQ:
                return "tool_calls"
            else:
                return "stop"
        case _:
            return "unknown"


# This function call parsing code is adapted from the anthropic-tools package (which is in "alpha"
# and not on PyPI, This will likely end up in the main anthropic package -- when that happens we'll
# switch to using that. Here is the commit we forked:
# https://github.com/anthropics/anthropic-tools/blob/a7822678db8a0867b1d05da9c836c456d263e3d9/tool_use_package/tool_user.py#L243


class FunctionCall:
    def __init__(self, function: str, parameters: list[tuple[str, str]]) -> None:
        self.function = function
        self.parameters = parameters


def message_content(message: Message) -> str:
    return "\n".join([content.text for content in message.content])


class ContentWithFunctionCalls:
    def __init__(
        self,
        content: str,
        function_calls: list[FunctionCall],
    ) -> None:
        self.content = content
        self.function_calls = function_calls


def extract_function_calls(message: Message) -> ContentWithFunctionCalls | None:
    content = message_content(message)

    # see if we need to append the </function_calls> stop token
    if (
        message.stop_reason == "stop_sequence"
        and message.stop_sequence == "</function_calls>"
    ):
        content = f"{content}</function_calls>"

    """Check if the function call follows a valid format and extract the attempted function calls if so.
    Does not check if the tools actually exist or if they are called with the requisite params."""
    # Check if there are any of the relevant XML tags present that would indicate an attempted function call.
    function_call_tags = re.findall(
        r"<function_calls>|</function_calls>|<invoke>|</invoke>|<tool_name>|</tool_name>|<parameters>|</parameters>",
        content,
        re.DOTALL,
    )
    if not function_call_tags:
        return None

    # Extract content between <function_calls> tags. If there are multiple we will only parse the first and ignore the rest, regardless of their correctness.
    match = re.search(r"<function_calls>(.*)</function_calls>", content, re.DOTALL)
    if not match:
        return None
    func_calls = match.group(1)

    # get content appearing before the function calls
    prefix_match = re.search(r"^(.*?)<function_calls>", content, re.DOTALL)
    if prefix_match:
        func_call_prefix_content = prefix_match.group(1)

    # Check for invoke tags
    invoke_regex = r"<invoke>.*?</invoke>"
    if not re.search(invoke_regex, func_calls, re.DOTALL):
        raise ValueError(
            "Missing <invoke></invoke> tags inside of <function_calls></function_calls> tags."
        )

    # Check each invoke contains tool name and parameters
    invoke_strings = re.findall(invoke_regex, func_calls, re.DOTALL)
    invokes: list[FunctionCall] = []
    for invoke_string in invoke_strings:
        tool_name = re.findall(r"<tool_name>.*?</tool_name>", invoke_string, re.DOTALL)
        if not tool_name:
            raise ValueError(
                "Missing <tool_name></tool_name> tags inside of <invoke></invoke> tags."
            )

        if len(tool_name) > 1:
            raise ValueError(
                "More than one tool_name specified inside single set of <invoke></invoke> tags."
            )

        parameters = re.findall(
            r"<parameters>.*?</parameters>", invoke_string, re.DOTALL
        )
        if not parameters:
            raise ValueError(
                "Missing <parameters></paraeters> tags inside of <invoke></invoke> tags."
            )

        if len(parameters) > 1:
            raise ValueError(
                "More than one set of <parameters></parameters> tags specified inside single set of <invoke></invoke> tags."
            )

        # Check for balanced tags inside parameters
        # TODO: This will fail if the parameter value contains <> pattern or if there is a parameter called parameters. Fix that issue.
        tags = re.findall(
            r"<.*?>",
            parameters[0].replace("<parameters>", "").replace("</parameters>", ""),
            re.DOTALL,
        )
        if len(tags) % 2 != 0:
            raise ValueError("Imbalanced tags inside <parameters></parameters> tags.")

        # Loop through the tags and check if each even-indexed tag matches the tag in the position after it (with the / of course).
        # If valid store their content for later use.
        # TODO: Add a check to make sure there aren't duplicates provided of a given parameter.
        parameters_with_values = []
        for i in range(0, len(tags), 2):
            opening_tag = tags[i]
            closing_tag = tags[i + 1]
            closing_tag_without_second_char = closing_tag[:1] + closing_tag[2:]
            if closing_tag[1] != "/" or opening_tag != closing_tag_without_second_char:
                raise ValueError(
                    "Non-matching opening and closing tags inside <parameters></parameters> tags."
                )

            match_param = re.search(
                rf"{opening_tag}(.*?){closing_tag}", parameters[0], re.DOTALL
            )
            if match_param:
                parameters_with_values.append((opening_tag[1:-1], match_param.group(1)))

        # Parse out the full function call
        invokes.append(
            FunctionCall(
                tool_name[0].replace("<tool_name>", "").replace("</tool_name>", ""),
                parameters_with_values,
            )
        )

    return ContentWithFunctionCalls(func_call_prefix_content, invokes)


#######################################################################################
# Thse functions deal with converting Anthropic <function_call> to our native ToolCall
#######################################################################################


def tool_call(invoke: FunctionCall, tools: list[ToolInfo]) -> ToolCall:
    tool_def = next((tool for tool in tools if invoke.function == tool.name), None)
    return ToolCall(
        id=invoke.function,
        function=invoke.function,
        arguments=tool_arguments(invoke.parameters, tool_def),
        type="function",
    )


def tool_arguments(
    params: list[tuple[str, str]], tool_info: ToolInfo | None
) -> dict[str, Any]:
    arguments: dict[str, Any] = dict()
    for param in params:
        # get params
        name, value = param

        # coerce type if we have a tool_def
        if tool_info:
            type_str = next(
                (param.type for param in tool_info.params if param.name == name), None
            )
            if type_str:
                value = tool_argument_value(value, type_str)

        arguments[name] = value

    return arguments


def tool_argument_value(value: Any, type_str: str) -> Any:
    """Convert a string value into its appropriate Python data type based on the provided type string.

    Arg:
        value: the value to convert
        type_str: the type to convert the value to
    Returns:
        The value converted into the requested type or the original value
        if the conversion failed.
    """
    type_str = json_type_to_python_type(type_str)
    if type_str in ("list", "dict"):
        return ast.literal_eval(value)
    type_class = getattr(builtins, type_str)
    try:
        return type_class(value)
    except ValueError:
        return value



import inspect
import logging
from copy import deepcopy
from typing import Any, Callable, TypeVar, cast

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_create,
    registry_info,
    registry_lookup,
    registry_name,
    registry_tag,
)
from inspect_ai.model import ModelName

from .task import Task

MODEL_PARAM = "model"

logger = logging.getLogger(__name__)


TaskType = TypeVar("TaskType", bound=Callable[..., Task])


def task_register(
    task: TaskType, name: str, attribs: dict[str, Any], params: list[str]
) -> TaskType:
    r"""Register a task.

    Args:
        task (TaskType):
            function that returns a Task or class
            deriving from Task
        name (str): Name of task
        attribs (dict[str,Any]): Attributes of task decorator
        params (list[str]): Task parameter names

    Returns:
        Task with registry attributes.
    """
    registry_add(
        task,
        RegistryInfo(
            type="task", name=name, metadata=dict(attribs=attribs, params=params)
        ),
    )
    return task


def task_create(name: str, model: ModelName, **kwargs: Any) -> Task:
    r"""Create a Task based on its registered name.

    Tasks can be a function that returns a Task or a
    class deriving from Task.

    Args:
        name (str): Name of task (Optional, defaults to object name)
        model (ModelName): Model name
        **kwargs (dict): Optional creation arguments for the task

    Returns:
        Task with registry info attribute
    """
    # bring in model arg (first deepcopy as we will mutate it)
    # add model to task_args
    kwargs = deepcopy(kwargs)
    kwargs[MODEL_PARAM] = model

    # match kwargs params to signature (warn if param not found)
    # (note that we always pass the 'model' param but tasks arne't
    # required to consume it, so we don't warn for 'model')
    task = registry_lookup("task", name)
    task_info = registry_info(task)
    task_params: list[str] = task_info.metadata["params"]
    task_args: dict[str, Any] = {}
    for param in kwargs.keys():
        if param in task_params:
            task_args[param] = kwargs[param]
        elif param != MODEL_PARAM:
            logger.warning(f"param '{param}' not used by task '{name}'")

    return cast(Task, registry_create("task", name, **task_args))


def task(*task: TaskType | None, name: str | None = None, **attribs: Any) -> Any:
    r"""Decorator for registering tasks.

    Args:
      *task (TaskType): Function returning `Task` targeted by
        plain task decorator without attributes (e.g. `@task`)
      name (str | None):
        Optional name for task. If the decorator has no name
        argument then the name of the function
        will be used to automatically assign a name.
      **attribs: (dict[str,Any]): Additional task attributes.

    Returns:
        Task with registry attributes.
    """

    def create_task_wrapper(task_type: TaskType) -> TaskType:
        # get the name and params
        task_name = registry_name(task_type, name or getattr(task_type, "__name__"))
        params = list(inspect.signature(task_type).parameters.keys())

        # create and return the wrapper
        def wrapper(*w_args: Any, **w_kwargs: Any) -> Task:
            # create the task
            task = task_type(*w_args, **w_kwargs)

            # tag it
            registry_tag(
                task_type,
                task,
                RegistryInfo(
                    type="task",
                    name=task_name,
                    metadata=dict(attribs=attribs, params=params),
                ),
                *w_args,
                **w_kwargs,
            )

            # return it
            return task

        return task_register(
            task=cast(TaskType, wrapper), name=task_name, attribs=attribs, params=params
        )

    if task:
        return create_task_wrapper(cast(TaskType, task[0]))
    else:
        return create_task_wrapper



from typing import Any, Callable, cast

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_name,
    registry_tag,
)

from ._model import ModelAPI


def modelapi_register(
    model_type: type[ModelAPI], name: str, models: list[str]
) -> type[ModelAPI]:
    r"""Register a model api.

    Args:
        model_type (type[Model]): Class deriving from Model
        name (str): API serving this model
        models (list[str]): Model names by this API

    Returns:
        Model API with registry attributes.
    """
    registry_add(
        model_type,
        RegistryInfo(type="modelapi", name=name, metadata=dict(models=models)),
    )
    return model_type


def modelapi(name: str, models: list[str] = []) -> Callable[..., type[ModelAPI]]:
    r"""Decorator for registering model APIs.

    Args:
        name (str): Name of API
        models (list[str]): Model names that should match this API.
          If no `models` are provided then this model type will always
          require an API prefix (e.g. "hf/openai-community/gpt2")

    Returns:
        Model API with registry attributes.
    """

    # create_model_wrapper:
    #  (a) Add the type[Model] to the registry using the appropriately
    #      package-namespaced name
    #  (b) Ensure that instances of Model created by type[Model] also
    #      carry registry info.
    def create_model_wrapper(
        wrapped: type[ModelAPI] | Callable[..., type[ModelAPI]], api: str
    ) -> type[ModelAPI]:
        model_api = registry_name(wrapped, api)

        def model_wrapper(*args: Any, **kwargs: Any) -> ModelAPI:
            if not isinstance(wrapped, type):
                model_type = wrapped()
            else:
                model_type = wrapped

            model = model_type(*args, **kwargs)
            registry_tag(
                model_type,
                model,
                RegistryInfo(
                    type="modelapi",
                    name=model_api,
                    metadata=dict(models=models),
                ),
                *args,
                **kwargs,
            )
            return model

        return modelapi_register(cast(type[ModelAPI], model_wrapper), model_api, models)

    def wrapper(
        model_type: type[ModelAPI] | Callable[..., type[ModelAPI]],
    ) -> type[ModelAPI]:
        return create_model_wrapper(model_type, name)

    return wrapper



from inspect_ai.model import ChatMessage, ChatMessageSystem


def append_system_message(
    messages: list[ChatMessage], message: ChatMessageSystem
) -> None:
    # find last index of any existing system message
    lastIndex = -1
    for i in list(reversed(range(0, len(messages)))):
        if isinstance(messages[i], ChatMessageSystem):
            lastIndex = i
            break

    # insert it
    messages.insert(lastIndex + 1, message)



from typing import Any

import yaml


def parse_cli_args(args: tuple[str] | list[str] | None) -> dict[str, Any]:
    params: dict[str, Any] = dict()
    if args:
        for arg in list(args):
            parts = arg.split("=")
            if len(parts) > 1:
                key = parts[0].replace("-", "_")
                value = yaml.safe_load("=".join(parts[1:]))
                if isinstance(value, str):
                    value = value.split(",")
                    value = value if len(value) > 1 else value[0]
                params[key] = value
    return params



"""
GPQA: A Graduate-Level Google-Proof Q&A Benchmark

David Rein, Betty Li Hou, Asa Cooper Stickland, Jackson Petty, Richard
Yuanzhe Pang, Julien Dirani, Julian Michael, Samuel R. Bowman
https://arxiv.org/abs/2311.12022

Based on: https://github.com/openai/simple-evals/blob/main/gpqa_eval.py

# eval for default epochs (4)
inspect eval gpqa.py

# eval with 1 epoch
inspect eval gpqa.py --epochs 1

# without chain of thought
inspect eval gpqa.py -T cot=false
"""


from inspect_ai import Task, task
from inspect_ai.dataset import Sample, csv_dataset
from inspect_ai.model import GenerateConfig
from inspect_ai.scorer import answer
from inspect_ai.solver import multiple_choice

# default epochs to run eval for
DEFAULT_EPOCHS = 4


# map records to inspect samples (note that target is always "A" in the,
# dataset, we will shuffle the presentation of options to mitigate this)
def record_to_sample(record):
    return Sample(
        input=record["Question"],
        choices=[
            str(record["Correct Answer"]),
            str(record["Incorrect Answer 1"]),
            str(record["Incorrect Answer 2"]),
            str(record["Incorrect Answer 3"]),
        ],
        target="A",
        id=record["Record ID"],
    )


@task
def gpqa_diamond(cot=True):
    return Task(
        dataset=csv_dataset(
            csv_file="https://openaipublic.blob.core.windows.net/simple-evals/gpqa_diamond.csv",
            sample_fields=record_to_sample,
        ),
        plan=[
            multiple_choice(cot=cot, shuffle=True),
        ],
        scorer=answer("letter"),
        config=GenerateConfig(temperature=0.5),
        epochs=DEFAULT_EPOCHS,
    )



import json
from typing import Any, Dict, Protocol, cast, runtime_checkable

from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.messages import ToolCall as LCToolCall
from langchain_core.outputs import (
    ChatGeneration,
    ChatResult,
)
from pydantic.v1 import Field
from typing_extensions import override

from inspect_ai.model import (
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
    Content,
    ContentImage,
    ContentText,
    GenerateConfig,
    ModelName,
    ModelOutput,
    ToolCall,
    ToolChoice,
    ToolInfo,
    ToolParam,
    get_model,
)
from inspect_ai.solver import Generate, Solver, TaskState


@runtime_checkable
class LangChainAgent(Protocol):
    async def __call__(
        self, llm: BaseChatModel, input: dict[str, Any]
    ) -> str | list[str | dict[str, Any]]:
        ...


def langchain_solver(agent: LangChainAgent) -> Solver:
    async def solve(state: TaskState, generate: Generate) -> TaskState:
        # create the inspect model api bridge
        llm = InspectChatModel()

        # call the agent
        await agent(
            llm=llm,
            input=dict(
                input=state.user_prompt.text,
                chat_history=as_langchain_chat_history(state.messages[1:]),
            ),
        )

        # collect output from llm interface
        state.messages = llm.messages
        state.output = llm.output

        # return state
        return state

    return solve


class InspectChatModel(BaseChatModel):
    # track messages and model output so we can update
    # the inspect task state when we are complete
    messages: list[ChatMessage] = Field(default=[], exclude=True)
    output: ModelOutput = Field(default=ModelOutput(), exclude=True)

    @property
    def _llm_type(self) -> str:
        return f"Inspect ({ModelName(get_model()).api})"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        return {
            "model_name": str(ModelName(get_model()).name),
        }

    @override
    def _generate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        # inspect uses async exclusively
        raise NotImplementedError

    @override
    async def _agenerate(
        self,
        messages: list[BaseMessage],
        stop: list[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        **kwargs: dict[str, Any],
    ) -> ChatResult:
        # extract tools from kwargs
        tools: list[ToolInfo] = []
        tool_choice: ToolChoice | None = None
        lc_tools = cast(list[dict[str, Any]] | None, kwargs.get("tools", None))
        if lc_tools:
            tools = [
                ToolInfo(
                    name=tool["function"]["name"],
                    description=tool["function"]["description"],
                    params=as_inspect_tool_params(tool["function"]["parameters"]),
                )
                for tool in lc_tools
            ]
            tool_choice = "auto"

        # generate
        input = [as_inspect_message(message) for message in messages]
        result = await get_model().generate(
            input=input,
            tools=tools,
            tool_choice=tool_choice,
            config=GenerateConfig(stop_seqs=stop),
        )

        # track last messages / model output
        self.messages = input
        self.messages.append(result.choices[0].message)
        self.output = result

        # extract choices
        generations = [
            ChatGeneration(message=as_langchain_message(choice.message))
            for choice in result.choices
        ]

        # return
        return ChatResult(generations=generations)


def as_inspect_message(message: BaseMessage) -> ChatMessage:
    if isinstance(message, SystemMessage):
        return ChatMessageSystem(content=as_inspect_content(message.content))
    elif isinstance(message, HumanMessage):
        return ChatMessageUser(content=as_inspect_content(message.content))
    elif isinstance(message, AIMessage):
        return ChatMessageAssistant(
            content=as_inspect_content(message.content),
            tool_calls=(
                [
                    ToolCall(
                        type="function",
                        function=call["name"],
                        id=call["id"] or call["name"],
                        arguments=call["args"],
                    )
                    for call in message.tool_calls
                ]
                if message.tool_calls and len(message.tool_calls) > 0
                else None
            ),
        )
    elif isinstance(message, ToolMessage):
        return ChatMessageTool(
            content=as_inspect_content(message.content),
            tool_call_id=message.tool_call_id,
        )
    elif isinstance(message, FunctionMessage):
        return ChatMessageTool(
            content=as_inspect_content(message.content), tool_call_id=message.name
        )
    else:
        raise ValueError(f"Unexpected message type: {type(message)}")


def as_langchain_message(message: ChatMessage) -> BaseMessage:
    if isinstance(message, ChatMessageSystem):
        return SystemMessage(content=as_langchain_content(message.content))
    elif isinstance(message, ChatMessageUser):
        return HumanMessage(content=as_langchain_content(message.content))
    elif isinstance(message, ChatMessageAssistant):
        additional_kwargs: dict[str, Any] = {}
        if message.tool_calls and len(message.tool_calls) > 0:
            additional_kwargs["tool_calls"] = [
                dict(
                    id=call.id, name=call.function, arguments=json.dumps(call.arguments)
                )
                for call in message.tool_calls
            ]

        return AIMessage(
            content=as_langchain_content(message.content),
            tool_calls=(
                [
                    LCToolCall(id=call.id, name=call.function, args=call.arguments)
                    for call in message.tool_calls
                ]
                if message.tool_calls
                else []
            ),
            additional_kwargs=additional_kwargs,
        )
    elif isinstance(message, ChatMessageTool):
        return ToolMessage(
            content=as_langchain_content(message.content),
            tool_call_id=message.tool_call_id or "",
        )
    else:
        raise ValueError(f"Unexpected message type: {type(message)}")


def as_langchain_chat_history(messages: list[ChatMessage]) -> list[dict[str, Any]]:
    return [dict(role=message.role, content=message.text) for message in messages]


def as_inspect_content(
    content: str | list[str | dict[str, Any]],
) -> str | list[Content]:
    if isinstance(content, str):
        return content
    else:
        return [
            (
                ContentText(text=c)
                if isinstance(c, str)
                else (
                    ContentText(text=c["text"])
                    if c["type"] == "text"
                    else ContentImage(image=c["image"])
                )
            )
            for c in content
        ]


def as_inspect_tool_params(parameters: dict[str, Any]) -> list[ToolParam]:
    params: list[ToolParam] = []
    for key, param in parameters["properties"].items():
        params.append(
            ToolParam(
                name=key,
                type=param["type"],
                description=param.get("description", param.get("title")),
                optional=key not in parameters["required"],
            )
        )
    return params


def as_langchain_content(
    content: str | list[Content],
) -> str | list[str | dict[str, Any]]:
    if isinstance(content, str):
        return content
    else:
        return [c if isinstance(c, str) else c.model_dump() for c in content]



import os
from typing import Any

import httpx
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_TOKENS
from inspect_ai.model import (
    ChatMessage,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
)
from inspect_ai.model._providers.util import model_base_url

from .._tool import ToolChoice, ToolInfo
from .._util import (
    chat_api_input,
    chat_api_request,
    is_chat_api_rate_limit,
)

# CloudFlare supported models:
# https://developers.cloudflare.com/workers-ai/models/#text-generation


class CloudFlareAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)
        self.account_id = os.getenv("CLOUDFLARE_ACCOUNT_ID")
        if not self.account_id:
            raise RuntimeError("CLOUDFLARE_ACCOUNT_ID environment variable not set")
        self.api_token = os.getenv("CLOUDFLARE_API_TOKEN")
        if not self.api_token:
            raise RuntimeError("CLOUDFLARE_API_TOKEN environment variable not set")
        self.client = httpx.AsyncClient()
        base_url = model_base_url(base_url, "CLOUDFLARE_BASE_URL")
        self.base_url = (
            base_url if base_url else "https://api.cloudflare.com/client/v4/accounts"
        )
        self.model_args = model_args

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # chat url
        chat_url = f"{self.base_url}/{self.account_id}/ai/run/@cf"

        # chat api input
        json: dict[str, Any] = dict(**self.model_args)
        if config.max_tokens is not None:
            json["max_tokens"] = config.max_tokens
        json["messages"] = chat_api_input(input)

        # make the call
        response = await chat_api_request(
            self.client,
            model_name=self.model_name,
            url=f"{chat_url}/{self.model_name}",
            headers={"Authorization": f"Bearer {self.api_token}"},
            json=json,
            config=config,
        )

        # handle response
        if response["success"]:
            return ModelOutput.from_content(
                model=self.model_name, content=response["result"]["response"]
            )
        else:
            error = str(response.get("errors", "Unknown"))
            raise RuntimeError(f"Error calling {self.model_name}: {error}")

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        return is_chat_api_rate_limit(ex)

    # cloudflare enforces rate limits by model for each account
    @override
    def connection_key(self) -> str:
        return f"{self.account_id}{self.model_name}"

    # cloudflare defaults to 256 max tokens, not enough for evals
    @override
    def max_tokens(self) -> int:
        return DEFAULT_MAX_TOKENS



import asyncio
import logging
import os
from pathlib import Path
from typing import Any

from shortuuid import uuid
from typing_extensions import Unpack

from inspect_ai._display.logger import init_logger
from inspect_ai._util.dotenv import init_dotenv
from inspect_ai._util.path import cwd_relative_path
from inspect_ai._util.platform import platform_init
from inspect_ai._util.registry import registry_lookup
from inspect_ai._view.view import view_notify_eval
from inspect_ai.log import EvalConfig, EvalLog, EvalLogInfo, read_eval_log
from inspect_ai.log._file import JSONRecorder
from inspect_ai.model import (
    GenerateConfig,
    GenerateConfigArgs,
    Model,
    get_model,
)
from inspect_ai.model._model import init_async_context_model
from inspect_ai.solver import Solver
from inspect_ai.util._context import init_async_context

from .loader import resolve_tasks
from .log import EvalLogger
from .task import Tasks, TaskSpec, task_file, task_run_dir

log = logging.getLogger(__name__)


def eval(
    tasks: Tasks,
    model: str | Model | None = None,
    model_base_url: str | None = None,
    model_args: dict[str, Any] = dict(),
    task_args: dict[str, Any] = dict(),
    plan: Solver | list[Solver] | None = None,
    log_level: str | None = None,
    log_dir: str | None = None,
    limit: int | tuple[int, int] | None = None,
    epochs: int | None = None,
    max_messages: int | None = None,
    max_subprocesses: int | None = None,
    log_samples: bool | None = None,
    log_images: bool | None = None,
    score: bool = True,
    **kwargs: Unpack[GenerateConfigArgs],
) -> list[EvalLog]:
    r"""Evaluate tasks using a Model.

    Args:
        tasks: (Tasks): Task(s) to evaluate. If None, attempt
            to evaluate a task in the current working directory
        model (str | Model | None): Model for evaluation. If not
            specified uses the current eval's model, or failing that
            the value of the INSPECT_EVAL_MODEL environment variable.
        model_base_url: (str | None): Base URL for communicating
            with the model API.
        model_args (dict[str,Any]): Model creation parameters
        task_args (dict[str,Any]): Task arguments
        plan (Solver | list[Solver] | None): Alternative plan
           for evaluating task(s). Optional (uses task plan by default).
        log_level (str | None): "debug", "http", "info", "warning", "error",
           or "critical" (defaults to "info")
        log_dir (str | None): Output path for logging results
           (defaults to file log in ./logs directory).
        limit (int | tuple[int, int] | None): Limit evaluated samples
            (defaults to all samples).
        epochs (int | None): Number of times to repeat evaluation of
            samples (defaults to 1)
        max_messages (int | None): Maximum number of messages to allow
           in a task conversation.
        max_subprocesses (int | None): Maximum number of subprocesses to
           run in parallel (default is os.cpu_count())
        log_samples: (bool | None): Log detailed samples and scores (defaults to True)
        log_images: (bool | None): Log base64 encoded version of images,
            even if specified as a filename or URL (defaults to True)
        score (bool): Score output (defaults to True)
        **kwargs (GenerateConfigArgs): Model generation options.

    Returns:
        List of EvalLog (one for each task)
    """
    # standard platform init for top level entry points
    platform_init()

    return asyncio.run(
        eval_async(
            tasks=tasks,
            model=model,
            model_base_url=model_base_url,
            model_args=model_args,
            task_args=task_args,
            plan=plan,
            log_level=log_level,
            log_dir=log_dir,
            limit=limit,
            epochs=epochs,
            max_messages=max_messages,
            max_subprocesses=max_subprocesses,
            log_samples=log_samples,
            log_images=log_images,
            score=score,
            **kwargs,
        )
    )


async def eval_async(
    tasks: Tasks,
    model: str | Model | None = None,
    model_base_url: str | None = None,
    model_args: dict[str, Any] = dict(),
    task_args: dict[str, Any] = dict(),
    plan: Solver | list[Solver] | None = None,
    log_level: str | None = None,
    log_dir: str | None = None,
    limit: int | tuple[int, int] | None = None,
    epochs: int | None = None,
    max_messages: int | None = None,
    max_subprocesses: int | None = None,
    log_samples: bool | None = None,
    log_images: bool | None = None,
    score: bool = True,
    **kwargs: Unpack[GenerateConfigArgs],
) -> list[EvalLog]:
    r"""Evaluate tasks using a Model (async).

    tasks: (Tasks): Task(s) to evaluate. If None, attempt
        to evaluate a task in the current working directory
    model (str | Model | None): Model for evaluation. If not
        specified uses the current eval's model, or failing that
        the value of the INSPECT_EVAL_MODEL environment variable.
    model_base_url: (str | None): Base URL for communicating
        with the model API.
    model_args (dict[str,Any]): Model creation parameters
    task_args (dict[str,Any]): Task arguments
    plan (Solver | list[Solver] | None): Alternative plan
        for evaluating task(s). Optional (uses task plan by default).
    log_level (str | None): "debug", "http", "info", "warning", "error",
        or "critical" (defaults to "info")
    log_dir (str | None): Output path for logging results
        (defaults to file log in ./logs directory).
    limit (int | tuple[int, int] | None): Limit evaluated samples
        (defaults to all samples).
    epochs (int | None): Number of times to repeat evaluation of
        samples (defaults to 1)
    max_messages (int | None): Maximum number of messages to allow
        in a task conversation.
    max_subprocesses (int | None): Maximum number of subprocesses to
        run in parallel (default is os.cpu_count())
    log_samples: (bool | None): Log detailed samples and scores (defaults to True)
    log_images: (bool | None): Log base64 encoded version of images,
        even if specified as a filename or URL (defaults to True)
    score (bool): Score output (defaults to True)
    **kwargs (GenerateConfigArgs): Model generation options.

    Returns:
        List of EvalLog (one for each task)
    """
    # Provide .env and log support bootstrap for notebooks and invoking
    # an eval as a plain Python script (as opposed to via inspect eval)
    init_dotenv()
    init_logger(log_level)

    # resolve model
    model = get_model(
        model=model,
        base_url=model_base_url,
        config=GenerateConfig(**kwargs),
        **model_args,
    )

    # init async context vars
    init_async_context(max_subprocesses)
    init_async_context_model(model)

    # if this is a TaskSpec then we are being spotted our id
    if isinstance(tasks, TaskSpec):
        task_id = tasks.id
        tasks = tasks.task
    else:
        task_id = None

    # resolve tasks
    eval_tasks = resolve_tasks(tasks, model, task_args)

    # warn and return empty string if we resovled no tasks
    if len(eval_tasks) == 0:
        log.warning("No inspect tasks were found at the specified paths.")
        return []

    # resolve recorder
    log_dir = log_dir if log_dir else os.environ.get("INSPECT_LOG_DIR", "./logs")
    log_dir = cwd_relative_path(log_dir)
    recorder = JSONRecorder(log_dir)

    # build task names and versions (include version if > 0)
    task_names: list[str] = [task.name for task in eval_tasks]
    task_versions: list[int] = [task.version for task in eval_tasks]

    # create config
    eval_config = EvalConfig(
        limit=limit,
        epochs=epochs,
        max_messages=max_messages,
        max_subprocesses=max_subprocesses,
        log_samples=log_samples,
        log_images=log_images,
    )

    run_id = uuid()
    loggers: list[EvalLogger] = []
    results: list[EvalLog] = []
    for index, name, version, task in zip(
        range(0, len(task_names)), task_names, task_versions, eval_tasks
    ):
        # tasks can provide their own epochs and max_messages
        task_eval_config = eval_config.model_copy()
        if task.epochs is not None:
            task_eval_config.epochs = task.epochs
        if task.max_messages is not None:
            task_eval_config.max_messages = task.max_messages

        # create and track the logger
        logger = EvalLogger(
            task_name=name,
            task_version=version,
            task_file=task_file(task, True),
            task_run_dir=task_run_dir(task),
            task_id=task_id if task_id else uuid(),
            run_id=run_id,
            model=model,
            dataset=task.dataset,
            task_attribs=task.attribs,
            task_args=task_args,
            model_args=model_args,
            eval_config=task_eval_config,
            recorder=recorder,
        )
        loggers.append(logger)

        # run the eval
        result = await task.run(
            sequence=(index + 1, len(task_names)),
            model=model,
            logger=logger,
            config=task_eval_config,
            plan=plan,
            score=score,
            **kwargs,
        )

        # mark completed and append result
        results.append(result)

        # notify the view module that an eval just completed
        # (in case we have a view polling for new evals)
        view_notify_eval(logger.location)

    # return list of eval logs
    return EvalLogs(results)


def eval_retry(
    tasks: EvalLogInfo | EvalLog | list[EvalLogInfo] | list[EvalLog],
    log_level: str | None = None,
    log_dir: str | None = None,
    max_subprocesses: int | None = None,
    log_samples: bool | None = None,
    log_images: bool | None = None,
    score: bool = True,
    max_retries: int | None = None,
    timeout: int | None = None,
    max_connections: int | None = None,
) -> list[EvalLog]:
    """Retry a previously failed evaluation task.

    Args:
        tasks: (EvalLogInfo | EvalLog | list[EvalLogInfo] | list[EvalLog]):
            Log files for task(s) to retry.
        log_level (str | None): "debug", "http", "info", "warning", "error",
           or "critical" (defaults to "info")
        log_dir (str | None): Output path for logging results
           (defaults to file log in ./logs directory).
        max_subprocesses (int | None): Maximum number of subprocesses to
           run in parallel (default is os.cpu_count())
        log_samples: (bool | None): Log detailed samples and scores (defaults to True)
        log_images: (bool | None): Log base64 encoded version of images,
           even if specified as a filename or URL (defaults to True)
        score (bool): Score output (defaults to True)
        max_retries (int | None):
           Maximum number of times to retry request.
        timeout: (int | None):
           Request timeout (in seconds)
        max_connections (int | None):
           Maximum number of concurrent connections to Model API (default is per Model API)

    Returns:
        List of EvalLog (one for each task)
    """
    platform_init()

    return asyncio.run(
        eval_retry_async(
            tasks=tasks,
            log_level=log_level,
            log_dir=log_dir,
            max_subprocesses=max_subprocesses,
            log_samples=log_samples,
            log_images=log_images,
            score=score,
            max_retries=max_retries,
            timeout=timeout,
            max_connections=max_connections,
        )
    )


async def eval_retry_async(
    tasks: EvalLogInfo | EvalLog | list[EvalLogInfo] | list[EvalLog],
    log_level: str | None = None,
    log_dir: str | None = None,
    max_subprocesses: int | None = None,
    log_samples: bool | None = None,
    log_images: bool | None = None,
    score: bool = True,
    max_retries: int | None = None,
    timeout: int | None = None,
    max_connections: int | None = None,
) -> list[EvalLog]:
    """Retry a previously failed evaluation task.

    Args:
        tasks: (EvalLogInfo | EvalLog | list[EvalLogInfo] | list[EvalLog]):
            Log files for task(s) to retry.
        log_level (str | None): "debug", "http", "info", "warning", "error",
           or "critical" (defaults to "info")
        log_dir (str | None): Output path for logging results
           (defaults to file log in ./logs directory).
        max_subprocesses (int): Maximum number of subprocesses to
           run in parallel (default is os.cpu_count())
        log_samples: (bool | None): Log detailed samples and scores (defaults to True)
        log_images: (bool | None): Log base64 encoded version of images,
           even if specified as a filename or URL (defaults to True)
        score (bool): Score output (defaults to True)
        max_retries (int | None):
           Maximum number of times to retry request.
        timeout: (int | None):
           Request timeout (in seconds)
        max_connections (int | None):
           Maximum number of concurrent connections to Model API (default is per Model API)

    Returns:
        List of EvalLog (one for each task)
    """
    # resolve into a list of eval logs
    if isinstance(tasks, EvalLogInfo):
        tasks = [tasks]
    elif isinstance(tasks, EvalLog):
        tasks = [tasks]
    retry_eval_logs = [
        task if isinstance(task, EvalLog) else read_eval_log(task.name)
        for task in tasks
    ]

    # eval them in turn
    eval_logs: list[EvalLog] = []
    for eval_log in retry_eval_logs:
        # the task needs to be either filesystem or registry
        # based in order to do a retry (we don't have enough
        # context to reconstruct ephemeral Task instances)
        task: str | None
        task_id = eval_log.eval.task_id
        task_name = eval_log.eval.task
        task_file = eval_log.eval.task_file
        if task_file:
            if not Path(task_file).exists():
                raise FileNotFoundError("Task file '{task_file}' not found")
            task = f"{task_file}@{task_name}"
        else:
            if registry_lookup("task", task_name) is None:
                raise FileNotFoundError("Task '{task_name}' not found.")
            task = task_name

        # collect the rest of the params we need for the eval
        model = eval_log.eval.model
        model_base_url = eval_log.eval.model_base_url
        model_args = eval_log.eval.model_args
        task_args = eval_log.eval.task_args
        limit = eval_log.eval.config.limit
        epochs = eval_log.eval.config.epochs
        max_messages = eval_log.eval.config.max_messages
        max_subprocesses = max_subprocesses or eval_log.eval.config.max_subprocesses
        log_samples = eval_log.eval.config.log_samples
        log_images = eval_log.eval.config.log_images
        config = eval_log.plan.config
        config.max_retries = max_retries or config.max_retries
        config.timeout = timeout or config.timeout
        config.max_connections = max_connections or config.max_connections

        # run the eval
        log = (
            await eval_async(
                tasks=TaskSpec(task=task, id=task_id),
                model=model,
                model_base_url=model_base_url,
                model_args=model_args,
                task_args=task_args,
                log_level=log_level,
                log_dir=log_dir,
                limit=limit,
                epochs=epochs,
                max_messages=max_messages,
                max_subprocesses=max_subprocesses,
                log_samples=log_samples,
                log_images=log_images,
                score=score,
                **dict(config),
            )
        )[0]

        # add it to our results
        eval_logs.append(log)

    return EvalLogs(eval_logs)


# A list of eval logs is returned from eval(). We've already displayed
# all of the ouptut we need to to though, so we make the return
# value 'invisible'
class EvalLogs(list[EvalLog]):
    def _ipython_display_(self) -> None:
        pass

    def __repr__(self) -> str:
        return ""



from typing import Any

from inspect_ai.model import ChatMessageSystem
from inspect_ai.util import resource

from ._solver import Generate, Solver, TaskState, solver
from ._util import append_system_message


@solver
def prompt_template(template: str, **params: dict[str, Any]) -> Solver:
    """Parameterized prompt template.

    Prompt template containing a `{prompt}` placeholder and any
    number of additional `params`.

    Args:
      template (str | list[Message]):
          The conversation template to use. A sipmle string or
          a list of messages
      **params (dict[str,Any]):
          A mapping of the parameters to fill into the template
          excluding the `{prompt}` parameter which is taken
          from the input.

    Returns:
      A solver that uses the specified prompt template.
    """
    # determine the prompt template
    prompt_template = resource(template)

    async def solve(state: TaskState, generate: Generate) -> TaskState:
        prompt = state.user_prompt
        prompt.text = prompt_template.format(prompt=prompt.text, **params)
        return state

    return solve


@solver
def system_message(message: str) -> Solver:
    """Solver which inserts a system message into the conversation.

    The new message will go after other system messages (if there
    are none it will be inserted at the beginnign of the conversation).

    Args:
       message (str): System message.
    """
    # read template
    content = resource(message)

    async def solve(state: TaskState, generate: Generate) -> TaskState:
        append_system_message(state.messages, ChatMessageSystem(content=content))
        return state

    return solve


DEFAULT_COT_TEMPLATE = r"""
{prompt}

Before answering, reason in a step-by-step manner as to get the right answer. Provide your answer at the end on its own line in the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the question.
"""


@solver
def chain_of_thought(template: str = DEFAULT_COT_TEMPLATE) -> Solver:
    """Solver which modifies the user prompt to encourage chain of thought.

    Args:
       template (str): String or path to file containing CoT template.
          The template uses a single variable: `prompt`.
    """
    return prompt_template(template)



from utils import run_example, skip_if_no_openai

from inspect_ai import Task, eval, score
from inspect_ai.dataset import Sample
from inspect_ai.scorer import Score, Scorer, Target, accuracy, includes, scorer
from inspect_ai.scorer._scorer import scorer_create
from inspect_ai.solver import TaskState


@scorer(metrics=[accuracy()], name="test_match")
def match() -> Scorer:
    async def score(state: TaskState, target: Target) -> Score:
        return (
            Score(value="C")
            if state.output.completion == target.text
            else Score(value="I")
        )

    return score


def test_scorer_lookup():
    scorer = scorer_create("test_match")
    assert scorer


@skip_if_no_openai
def test_no_scorer():
    task = Task(
        dataset=[Sample(input="What is 1 + 1?", target=["2", "2.0", "Two"])],
    )
    log = eval(task)[0]
    assert log.samples[0].score is None


@skip_if_no_openai
def test_score_function():
    log = run_example("popularity.py", "openai/gpt-4")
    log = score(log[0], includes())
    assert log.samples[0].score.value
